# SPDX-License-Identifier: Apache-2.0


from ..proto import onnx_proto
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_feature_selection(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    op = operator.raw_operator
    # Get indices of the features selected
    index = op.get_support(indices=True)
    if len(index) == 0:
        raise RuntimeError(
            "Model '{}' did not select any feature. "
            "This model cannot be converted into ONNX."
            "".format(op.__class__.__name__)
        )
    output_name = operator.outputs[0].full_name
    if index.any():
        column_indices_name = scope.get_unique_variable_name("column_indices")

        container.add_initializer(
            column_indices_name, onnx_proto.TensorProto.INT64, [len(index)], index
        )

        container.add_node(
            "ArrayFeatureExtractor",
            [operator.inputs[0].full_name, column_indices_name],
            output_name,
            op_domain="ai.onnx.ml",
            name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
        )
    else:
        container.add_node(
            "ConstantOfShape", operator.inputs[0].full_name, output_name, op_version=9
        )


register_converter("SklearnGenericUnivariateSelect", convert_sklearn_feature_selection)
register_converter("SklearnRFE", convert_sklearn_feature_selection)
register_converter("SklearnRFECV", convert_sklearn_feature_selection)
register_converter("SklearnSelectFdr", convert_sklearn_feature_selection)
register_converter("SklearnSelectFpr", convert_sklearn_feature_selection)
register_converter("SklearnSelectFromModel", convert_sklearn_feature_selection)
register_converter("SklearnSelectFwe", convert_sklearn_feature_selection)
register_converter("SklearnSelectKBest", convert_sklearn_feature_selection)
register_converter("SklearnSelectPercentile", convert_sklearn_feature_selection)
register_converter("SklearnVarianceThreshold", convert_sklearn_feature_selection)



# SPDX-License-Identifier: Apache-2.0

from ..proto import onnx_proto
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_sequence_at(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    i_index = operator.index
    index_name = scope.get_unique_variable_name("seq_at%d" % i_index)
    container.add_initializer(index_name, onnx_proto.TensorProto.INT64, [], [i_index])
    container.add_node(
        "SequenceAt",
        [operator.inputs[0].full_name, index_name],
        operator.outputs[0].full_name,
        name=scope.get_unique_operator_name("SequenceAt%d" % i_index),
    )


def convert_sklearn_sequence_construct(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    container.add_node(
        "SequenceConstruct",
        [i.full_name for i in operator.inputs],
        operator.outputs[0].full_name,
        name=scope.get_unique_operator_name("SequenceConstruct"),
    )


register_converter("SklearnSequenceAt", convert_sklearn_sequence_at)
register_converter("SklearnSequenceConstruct", convert_sklearn_sequence_construct)



# SPDX-License-Identifier: Apache-2.0

from sklearn.base import is_regressor
from ..proto import onnx_proto
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._apply_operation import apply_cast, apply_concat, apply_reshape
from ..common.data_types import guess_proto_type, Int64TensorType
from .._supported_operators import sklearn_operator_name_map


def _iteration_one_versus(
    scope,
    container,
    inputs,
    i,
    estimator,
    cl_type,
    proto_dtype,
    use_raw_scores=True,
    prob_shape=None,
):
    op_type = sklearn_operator_name_map[type(estimator)]

    this_operator = scope.declare_local_operator(op_type, raw_model=estimator)
    this_operator.inputs = inputs

    if is_regressor(estimator):
        score_name = scope.declare_local_variable("score_%d" % i, cl_type())
        this_operator.outputs.append(score_name)

        if hasattr(estimator, "coef_") and len(estimator.coef_.shape) == 2:
            raise RuntimeError(
                "OneVsRestClassifier or OneVsOneClassifier accepts "
                "regressor with only one target."
            )
        p1 = score_name.onnx_name
        return None, None, p1

    if container.has_options(estimator, "raw_scores"):
        options = {"raw_scores": use_raw_scores}
    elif container.has_options(estimator, "zipmap"):
        options = {"zipmap": False}
    else:
        options = None
    if options is not None:
        container.add_options(id(estimator), options)
        scope.add_options(id(estimator), options)

    label_name = scope.declare_local_variable("label_%d" % i, Int64TensorType())
    prob_name = scope.declare_local_variable("proba_%d" % i, inputs[0].type.__class__())
    this_operator.outputs.append(label_name)
    this_operator.outputs.append(prob_name)

    # gets the label for the class 1
    label = scope.get_unique_variable_name("lab_%d" % i)
    apply_reshape(scope, label_name.onnx_name, label, container, desired_shape=(-1, 1))
    cast_label = scope.get_unique_variable_name("cast_lab_%d" % i)
    apply_cast(scope, label, cast_label, container, to=proto_dtype)

    # get the probability for the class 1
    if prob_shape is None:
        # shape to use to reshape score
        cst0 = scope.get_unique_variable_name("cst0")
        container.add_initializer(cst0, onnx_proto.TensorProto.INT64, [1], [0])
        shape = scope.get_unique_variable_name("shape")
        container.add_node("Shape", [inputs[0].full_name], [shape])
        first_dim = scope.get_unique_variable_name("dim")
        container.add_node("Gather", [shape, cst0], [first_dim])
        cst_1 = scope.get_unique_variable_name("cst_1")
        container.add_initializer(cst_1, onnx_proto.TensorProto.INT64, [1], [-1])
        prob_shape = scope.get_unique_variable_name("shape")
        apply_concat(scope, [first_dim, cst_1], prob_shape, container, axis=0)

    prob_reshaped = scope.get_unique_variable_name("prob_%d" % i)
    container.add_node("Reshape", [prob_name.onnx_name, prob_shape], [prob_reshaped])

    cst1 = scope.get_unique_variable_name("cst1")
    container.add_initializer(cst1, onnx_proto.TensorProto.INT64, [1], [1])
    cst2 = scope.get_unique_variable_name("cst2")
    container.add_initializer(cst2, onnx_proto.TensorProto.INT64, [1], [2])

    prob1 = scope.get_unique_variable_name("prob1_%d" % i)
    container.add_node("Slice", [prob_reshaped, cst1, cst2, cst1], prob1)
    return prob_shape, cast_label, prob1


def convert_one_vs_one_classifier(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT
    op = operator.raw_operator

    # shape to use to reshape score
    cst0 = scope.get_unique_variable_name("cst0")
    container.add_initializer(cst0, onnx_proto.TensorProto.INT64, [1], [0])
    cst1 = scope.get_unique_variable_name("cst1")
    container.add_initializer(cst1, onnx_proto.TensorProto.INT64, [1], [1])
    cst2 = scope.get_unique_variable_name("cst2")
    container.add_initializer(cst2, onnx_proto.TensorProto.INT64, [1], [2])
    shape = scope.get_unique_variable_name("shape")
    container.add_node("Shape", [operator.inputs[0].full_name], [shape])
    first_dim = scope.get_unique_variable_name("dim")
    container.add_node("Gather", [shape, cst0], [first_dim])
    cst_1 = scope.get_unique_variable_name("cst_1")
    container.add_initializer(cst_1, onnx_proto.TensorProto.INT64, [1], [-1])
    prob_shape = scope.get_unique_variable_name("shape")
    apply_concat(scope, [first_dim, cst_1], prob_shape, container, axis=0)

    label_names = []
    prob_names = []
    prob_shape = None
    cl_type = operator.inputs[0].type.__class__
    for i, estimator in enumerate(op.estimators_):
        prob_shape, cast_label, prob1 = _iteration_one_versus(
            scope,
            container,
            operator.inputs,
            i,
            estimator,
            cl_type,
            proto_dtype,
            True,
            prob_shape=prob_shape,
        )

        label_names.append(cast_label)
        prob_names.append(prob1)

    conc_lab_name = scope.get_unique_variable_name("concat_out_ovo_label")
    apply_concat(scope, label_names, conc_lab_name, container, axis=1)
    conc_prob_name = scope.get_unique_variable_name("concat_out_ovo_prob")
    apply_concat(scope, prob_names, conc_prob_name, container, axis=1)

    # calls _ovr_decision_function
    this_operator = scope.declare_local_operator("SklearnOVRDecisionFunction", op)

    cl_type = operator.inputs[0].type.__class__
    label = scope.declare_local_variable("label", cl_type())
    container.add_node("Identity", [conc_lab_name], [label.onnx_name])
    prob_score = scope.declare_local_variable("prob_score", cl_type())
    container.add_node("Identity", [conc_prob_name], [prob_score.onnx_name])

    this_operator.inputs.append(label)
    this_operator.inputs.append(prob_score)

    ovr_name = scope.declare_local_variable("ovr_output", cl_type())
    this_operator.outputs.append(ovr_name)

    output_name = operator.outputs[1].full_name
    container.add_node("Identity", [ovr_name.onnx_name], [output_name])

    container.add_node("ArgMax", "ovr_output", operator.outputs[0].full_name, axis=1)


register_converter(
    "SklearnOneVsOneClassifier",
    convert_one_vs_one_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "nocl": [True, False],
        "output_class_labels": [False, True],
    },
)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.data_types import Int64TensorType, StringTensorType
from ..common.utils import check_input_and_output_numbers
from ..common.utils import check_input_and_output_types


def calculate_sklearn_label_binariser_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[Int64TensorType, StringTensorType]
    )

    N = operator.inputs[0].get_first_dimension()
    operator.outputs[0].type = Int64TensorType([N, len(operator.raw_operator.classes_)])


register_shape_calculator(
    "SklearnLabelBinarizer", calculate_sklearn_label_binariser_output_shapes
)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..proto import onnx_proto
from ..common._apply_operation import apply_cast, apply_reshape
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_label_binariser(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """Converts Scikit Label Binariser model to onnx format."""
    binariser_op = operator.raw_operator
    classes = binariser_op.classes_
    if hasattr(binariser_op, "sparse_input_") and binariser_op.sparse_input_:
        raise RuntimeError("sparse is not supported for LabelBinarizer.")
    if (
        hasattr(binariser_op, "y_type_")
        and binariser_op.y_type_ == "multilabel-indicator"
    ):
        if binariser_op.pos_label != 1:
            raise RuntimeError("pos_label != 1 is not supported " "for LabelBinarizer.")
        if list(classes) != list(range(len(classes))):
            raise RuntimeError(
                "classes != [0, 1, ..., n_classes] is not "
                "supported for LabelBinarizer."
            )
        container.add_node(
            "Identity",
            operator.inputs[0].full_name,
            operator.output_full_names,
            name=scope.get_unique_operator_name("identity"),
        )
    else:
        zeros_tensor = np.full(
            (1, len(classes)), binariser_op.neg_label, dtype=np.float32
        )
        unit_tensor = np.full(
            (1, len(classes)), binariser_op.pos_label, dtype=np.float32
        )

        classes_tensor_name = scope.get_unique_variable_name("classes_tensor")
        equal_condition_tensor_name = scope.get_unique_variable_name(
            "equal_condition_tensor"
        )
        zeros_tensor_name = scope.get_unique_variable_name("zero_tensor")
        unit_tensor_name = scope.get_unique_variable_name("unit_tensor")
        where_result_name = scope.get_unique_variable_name("where_result")

        class_dtype = onnx_proto.TensorProto.STRING
        if (
            np.issubdtype(binariser_op.classes_.dtype, np.signedinteger)
            or binariser_op.classes_.dtype == np.bool_
        ):
            class_dtype = onnx_proto.TensorProto.INT64
        else:
            classes = np.array([s.encode("utf-8") for s in classes])

        container.add_initializer(
            classes_tensor_name, class_dtype, [len(classes)], classes
        )
        container.add_initializer(
            zeros_tensor_name,
            onnx_proto.TensorProto.FLOAT,
            zeros_tensor.shape,
            zeros_tensor.ravel(),
        )
        container.add_initializer(
            unit_tensor_name,
            onnx_proto.TensorProto.FLOAT,
            unit_tensor.shape,
            unit_tensor.ravel(),
        )

        reshaped_input_name = scope.get_unique_variable_name("reshaped_input")
        apply_reshape(
            scope,
            operator.inputs[0].full_name,
            reshaped_input_name,
            container,
            desired_shape=[-1, 1],
        )

        # Models with classes_/inputs of string type would fail in the
        # following step as Equal op does not support string comparison.
        container.add_node(
            "Equal",
            [classes_tensor_name, reshaped_input_name],
            equal_condition_tensor_name,
            name=scope.get_unique_operator_name("equal"),
        )
        container.add_node(
            "Where",
            [equal_condition_tensor_name, unit_tensor_name, zeros_tensor_name],
            where_result_name,
            name=scope.get_unique_operator_name("where"),
        )
        where_res = where_result_name

        if len(binariser_op.classes_) == 2:
            array_f_name = scope.get_unique_variable_name(
                "array_feature_extractor_result"
            )
            pos_class_index_name = scope.get_unique_variable_name("pos_class_index")

            container.add_initializer(
                pos_class_index_name, onnx_proto.TensorProto.INT64, [], [1]
            )

            container.add_node(
                "ArrayFeatureExtractor",
                [where_result_name, pos_class_index_name],
                array_f_name,
                op_domain="ai.onnx.ml",
                name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
            )
            where_res = array_f_name
        apply_cast(
            scope,
            where_res,
            operator.output_full_names,
            container,
            to=onnx_proto.TensorProto.INT64,
        )


register_converter("SklearnLabelBinarizer", convert_sklearn_label_binariser)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.data_types import FloatTensorType, Int64TensorType, DoubleTensorType
from ..common.utils import check_input_and_output_numbers
from ..common.utils import check_input_and_output_types


def calculate_sklearn_truncated_svd_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C] ---> [N, K]

    Transform feature dimension from C to K
    """
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(
        operator,
        good_input_types=[FloatTensorType, Int64TensorType, DoubleTensorType],
        good_output_types=[FloatTensorType, DoubleTensorType],
    )

    if len(operator.inputs[0].type.shape) != 2:
        raise RuntimeError("Only 2-D tensor(s) can be input(s).")

    cls_type = operator.inputs[0].type.__class__
    if cls_type != DoubleTensorType:
        cls_type = FloatTensorType
    N = operator.inputs[0].get_first_dimension()
    K = (
        operator.raw_operator.n_components
        if operator.type == "SklearnTruncatedSVD"
        else operator.raw_operator.n_components_
    )

    operator.outputs[0].type = cls_type([N, K])


register_shape_calculator(
    "SklearnIncrementalPCA", calculate_sklearn_truncated_svd_output_shapes
)
register_shape_calculator("SklearnPCA", calculate_sklearn_truncated_svd_output_shapes)
register_shape_calculator(
    "SklearnTruncatedSVD", calculate_sklearn_truncated_svd_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

from ..common._registration import register_shape_calculator
from ..common.data_types import Int64TensorType


def calculate_sgd_oneclass_svm_output_shapes(operator):
    N = operator.inputs[0].get_first_dimension()
    operator.outputs[0].type = Int64TensorType(
        [
            N,
        ]
    )
    operator.outputs[1].type.shape = [
        N,
    ]


register_shape_calculator(
    "SklearnSGDOneClassSVM", calculate_sgd_oneclass_svm_output_shapes
)



# SPDX-License-Identifier: Apache-2.0


import numbers
import numpy as np
from sklearn.ensemble import RandomTreesEmbedding
from ..common._apply_operation import (
    apply_cast,
    apply_concat,
    apply_reshape,
    apply_transpose,
)
from ..common.data_types import BooleanTensorType, Int64TensorType, guess_numpy_type
from ..common._registration import register_converter
from ..common.tree_ensemble import (
    add_tree_to_attribute_pairs,
    add_tree_to_attribute_pairs_hist_gradient_boosting,
    get_default_tree_classifier_attribute_pairs,
    get_default_tree_regressor_attribute_pairs,
)
from ..common.utils_classifier import get_label_classes
from ..proto import onnx_proto
from .decision_tree import (
    predict,
    _build_labels_path,
    _build_labels_leaf,
    _append_decision_output,
)


def _num_estimators(op):
    # don't use op.n_estimators since it may not be the same as
    # len(op.estimators_). At training time n_estimators can be changed by
    # training code:
    #   for j in range(10):
    #       ...
    #       classifier.fit(X_tmp, y_tmp)
    #       classifier.n_estimators += 30
    if hasattr(op, "estimators_"):
        return len(op.estimators_)
    elif hasattr(op, "_predictors"):
        # HistGradientBoosting*
        return len(op._predictors)
    raise NotImplementedError(
        "Model should have attribute 'estimators_' or '_predictors'."
    )


def _calculate_labels(scope, container, model, proba):
    predictions = []
    transposed_result_name = scope.get_unique_variable_name("transposed_result")
    apply_transpose(scope, proba, transposed_result_name, container, perm=(1, 2, 0))
    for k in range(model.n_outputs_):
        preds_name = scope.get_unique_variable_name("preds")
        reshaped_preds_name = scope.get_unique_variable_name("reshaped_preds")
        k_name = scope.get_unique_variable_name("k_column")
        out_k_name = scope.get_unique_variable_name("out_k_column")
        argmax_output_name = scope.get_unique_variable_name("argmax_output")
        classes_name = scope.get_unique_variable_name("classes")
        reshaped_result_name = scope.get_unique_variable_name("reshaped_result")

        container.add_initializer(k_name, onnx_proto.TensorProto.INT64, [], [k])
        container.add_initializer(
            classes_name,
            onnx_proto.TensorProto.INT64,
            model.classes_[k].shape,
            model.classes_[k],
        )

        container.add_node(
            "ArrayFeatureExtractor",
            [transposed_result_name, k_name],
            out_k_name,
            op_domain="ai.onnx.ml",
            name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
        )
        container.add_node(
            "ArgMax",
            out_k_name,
            argmax_output_name,
            name=scope.get_unique_operator_name("ArgMax"),
            axis=1,
        )
        apply_reshape(
            scope,
            argmax_output_name,
            reshaped_result_name,
            container,
            desired_shape=(1, -1),
        )
        container.add_node(
            "ArrayFeatureExtractor",
            [classes_name, reshaped_result_name],
            preds_name,
            op_domain="ai.onnx.ml",
            name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
        )
        apply_reshape(
            scope, preds_name, reshaped_preds_name, container, desired_shape=(-1, 1)
        )
        predictions.append(reshaped_preds_name)
    return predictions


def convert_sklearn_random_forest_classifier(
    scope,
    operator,
    container,
    op_type="TreeEnsembleClassifier",
    op_domain="ai.onnx.ml",
    op_version=1,
):
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    attr_dtype = dtype if op_version >= 3 else np.float32
    op = operator.raw_operator

    if hasattr(op, "n_outputs_"):
        n_outputs = int(op.n_outputs_)
        options = container.get_options(
            op, dict(raw_scores=False, decision_path=False, decision_leaf=False)
        )
    elif hasattr(op, "n_trees_per_iteration_"):
        # HistGradientBoostingClassifier
        n_outputs = op.n_trees_per_iteration_
        options = container.get_options(op, dict(raw_scores=False))
    else:
        raise NotImplementedError(
            "Model should have attribute 'n_outputs_' or " "'n_trees_per_iteration_'."
        )

    use_raw_scores = options["raw_scores"]

    if n_outputs == 1 or hasattr(op, "loss_") or hasattr(op, "_loss"):
        classes = get_label_classes(scope, op)

        if all(isinstance(i, np.ndarray) for i in classes):
            classes = np.concatenate(classes)
        attr_pairs = get_default_tree_classifier_attribute_pairs()
        attr_pairs["name"] = scope.get_unique_operator_name(op_type)

        if all(isinstance(i, (numbers.Real, bool, np.bool_)) for i in classes):
            class_labels = [int(i) for i in classes]
            attr_pairs["classlabels_int64s"] = class_labels
        elif all(isinstance(i, str) for i in classes):
            class_labels = [str(i) for i in classes]
            attr_pairs["classlabels_strings"] = class_labels
        else:
            raise ValueError("Only string and integer class labels are allowed.")

        # random forest calculate the final score by averaging over all trees'
        # outcomes, so all trees' weights are identical.
        if hasattr(op, "estimators_"):
            estimator_count = len(op.estimators_)
            tree_weight = 1.0 / estimator_count
        elif hasattr(op, "_predictors"):
            # HistGradientBoostingRegressor
            estimator_count = len(op._predictors)
            tree_weight = 1.0
        else:
            raise NotImplementedError(
                "Model should have attribute 'estimators_' or '_predictors'."
            )

        for tree_id in range(estimator_count):
            if hasattr(op, "estimators_"):
                tree = op.estimators_[tree_id].tree_
                add_tree_to_attribute_pairs(
                    attr_pairs,
                    True,
                    tree,
                    tree_id,
                    tree_weight,
                    0,
                    True,
                    True,
                    dtype=dtype,
                )
            else:
                # HistGradientBoostClassifier
                if len(op._predictors[tree_id]) == 1:
                    tree = op._predictors[tree_id][0]
                    add_tree_to_attribute_pairs_hist_gradient_boosting(
                        attr_pairs,
                        True,
                        tree,
                        tree_id,
                        tree_weight,
                        0,
                        False,
                        False,
                        dtype=dtype,
                    )
                else:
                    for cl, tree in enumerate(op._predictors[tree_id]):
                        add_tree_to_attribute_pairs_hist_gradient_boosting(
                            attr_pairs,
                            True,
                            tree,
                            tree_id * n_outputs + cl,
                            tree_weight,
                            cl,
                            False,
                            False,
                            dtype=dtype,
                        )

        if hasattr(op, "_baseline_prediction"):
            if isinstance(op._baseline_prediction, np.ndarray):
                attr_pairs["base_values"] = list(op._baseline_prediction.ravel())
            else:
                attr_pairs["base_values"] = [op._baseline_prediction]

        if hasattr(op, "loss_"):
            loss = op.loss_
        elif hasattr(op, "_loss"):
            # scikit-learn >= 0.24
            loss = op._loss
        else:
            loss = None
        if loss is not None:
            if use_raw_scores:
                attr_pairs["post_transform"] = "NONE"
            elif loss.__class__.__name__ in ("BinaryCrossEntropy", "HalfBinomialLoss"):
                attr_pairs["post_transform"] = "LOGISTIC"
            elif loss.__class__.__name__ in (
                "CategoricalCrossEntropy",
                "HalfMultinomialLoss",
            ):
                attr_pairs["post_transform"] = "SOFTMAX"
            else:
                raise NotImplementedError(
                    "There is no corresponding post_transform for "
                    "'{}'.".format(loss.__class__.__name__)
                )
        elif use_raw_scores:
            raise RuntimeError(
                "The converter cannot implement decision_function for "
                "'{}' and loss '{}'.".format(type(op), loss)
            )

        input_name = operator.input_full_names
        if isinstance(operator.inputs[0].type, BooleanTensorType):
            cast_input_name = scope.get_unique_variable_name("cast_input")

            apply_cast(
                scope,
                input_name,
                cast_input_name,
                container,
                to=onnx_proto.TensorProto.FLOAT,
            )
            input_name = cast_input_name

        if dtype is not None:
            for k in attr_pairs:
                if k in (
                    "nodes_values",
                    "class_weights",
                    "target_weights",
                    "nodes_hitrates",
                    "base_values",
                ):
                    attr_pairs[k] = np.array(attr_pairs[k], dtype=attr_dtype).ravel()

        container.add_node(
            op_type,
            input_name,
            [operator.outputs[0].full_name, operator.outputs[1].full_name],
            op_domain=op_domain,
            op_version=op_version,
            **attr_pairs
        )

        if not options.get("decision_path", False) and not options.get(
            "decision_leaf", False
        ):
            return

        # decision_path
        tree_paths = []
        tree_leaves = []
        for i, tree in enumerate(op.estimators_):
            attrs = get_default_tree_classifier_attribute_pairs()
            attrs["name"] = scope.get_unique_operator_name("%s_%d" % (op_type, i))
            attrs["n_targets"] = int(op.n_outputs_)
            add_tree_to_attribute_pairs(
                attrs, True, tree.tree_, 0, 1.0, 0, False, True, dtype=dtype
            )

            attrs["n_targets"] = 1
            attrs["post_transform"] = "NONE"
            attrs["target_ids"] = [0 for _ in attrs["class_ids"]]
            attrs["target_weights"] = [float(_) for _ in attrs["class_nodeids"]]
            attrs["target_nodeids"] = attrs["class_nodeids"]
            attrs["target_treeids"] = attrs["class_treeids"]
            rem = [k for k in attrs if k.startswith("class")]
            for k in rem:
                del attrs[k]

            if dtype is not None:
                for k in attrs:
                    if k in (
                        "nodes_values",
                        "class_weights",
                        "target_weights",
                        "nodes_hitrates",
                        "base_values",
                    ):
                        attrs[k] = np.array(attrs[k], dtype=attr_dtype).ravel()

            if options["decision_path"]:
                # decision_path
                tree_paths.append(
                    _append_decision_output(
                        input_name,
                        attrs,
                        _build_labels_path,
                        None,
                        scope,
                        operator,
                        container,
                        op_type=op_type,
                        op_domain=op_domain,
                        op_version=op_version,
                        regression=True,
                        overwrite_tree=tree.tree_,
                    )
                )
            if options["decision_leaf"]:
                # decision_path
                tree_leaves.append(
                    _append_decision_output(
                        input_name,
                        attrs,
                        _build_labels_leaf,
                        None,
                        scope,
                        operator,
                        container,
                        op_type=op_type,
                        op_domain=op_domain,
                        op_version=op_version,
                        regression=True,
                        cast_encode=True,
                    )
                )

        # merges everything
        n_out = 2
        if options["decision_path"]:
            apply_concat(
                scope,
                tree_paths,
                operator.outputs[n_out].full_name,
                container,
                axis=1,
                operator_name=scope.get_unique_operator_name("concat"),
            )
            n_out += 1

        if options["decision_leaf"]:
            # decision_path
            apply_concat(
                scope,
                tree_leaves,
                operator.outputs[n_out].full_name,
                container,
                axis=1,
                operator_name=scope.get_unique_operator_name("concat"),
            )
            n_out += 1

    else:
        if use_raw_scores:
            raise RuntimeError(
                "The converter cannot implement decision_function for "
                "'{}'.".format(type(op))
            )
        concatenated_proba_name = scope.get_unique_variable_name("concatenated_proba")
        proba = []
        for est in op.estimators_:
            reshaped_est_proba_name = scope.get_unique_variable_name(
                "reshaped_est_proba"
            )
            est_proba = predict(
                est,
                scope,
                operator,
                container,
                op_type,
                op_domain,
                op_version,
                is_ensemble=True,
            )
            apply_reshape(
                scope,
                est_proba,
                reshaped_est_proba_name,
                container,
                desired_shape=(1, n_outputs, -1, max([len(x) for x in op.classes_])),
            )
            proba.append(reshaped_est_proba_name)
        apply_concat(scope, proba, concatenated_proba_name, container, axis=0)
        if container.target_opset >= 18:
            axis_name = scope.get_unique_variable_name("axis")
            container.add_initializer(axis_name, onnx_proto.TensorProto.INT64, [1], [0])
            container.add_node(
                "ReduceMean",
                [concatenated_proba_name, axis_name],
                operator.outputs[1].full_name,
                name=scope.get_unique_operator_name("ReduceMean"),
                keepdims=0,
            )
        else:
            container.add_node(
                "ReduceMean",
                concatenated_proba_name,
                operator.outputs[1].full_name,
                name=scope.get_unique_operator_name("ReduceMean"),
                axes=[0],
                keepdims=0,
            )
        predictions = _calculate_labels(
            scope, container, op, operator.outputs[1].full_name
        )
        apply_concat(
            scope, predictions, operator.outputs[0].full_name, container, axis=1
        )

        if options.get("decision_path", False) or options.get("decision_leaf", False):
            raise RuntimeError(
                "Decision output for multi-outputs is not implemented yet."
            )


def convert_sklearn_random_forest_regressor_converter(
    scope,
    operator,
    container,
    op_type="TreeEnsembleRegressor",
    op_domain="ai.onnx.ml",
    op_version=1,
):
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    op = operator.raw_operator
    attrs = get_default_tree_regressor_attribute_pairs()
    attrs["name"] = scope.get_unique_operator_name(op_type)

    if hasattr(op, "n_outputs_"):
        attrs["n_targets"] = int(op.n_outputs_)
    elif hasattr(op, "n_trees_per_iteration_"):
        # HistGradientBoostingRegressor
        attrs["n_targets"] = op.n_trees_per_iteration_
    else:
        raise NotImplementedError(
            "Model should have attribute 'n_outputs_' or " "'n_trees_per_iteration_'."
        )

    if hasattr(op, "estimators_"):
        estimator_count = len(op.estimators_)
        tree_weight = 1.0 / estimator_count
    elif hasattr(op, "_predictors"):
        # HistGradientBoostingRegressor
        estimator_count = len(op._predictors)
        tree_weight = 1.0
    else:
        raise NotImplementedError(
            "Model should have attribute 'estimators_' or '_predictors'."
        )

    # random forest calculate the final score by averaging over all trees'
    # outcomes, so all trees' weights are identical.
    for tree_id in range(estimator_count):
        if hasattr(op, "estimators_"):
            tree = op.estimators_[tree_id].tree_
            add_tree_to_attribute_pairs(
                attrs, False, tree, tree_id, tree_weight, 0, False, True, dtype=dtype
            )
        else:
            # HistGradientBoostingRegressor
            if len(op._predictors[tree_id]) != 1:
                raise NotImplementedError(
                    "The converter does not work when the number of trees "
                    "is not 1 but {}.".format(len(op._predictors[tree_id]))
                )
            tree = op._predictors[tree_id][0]
            add_tree_to_attribute_pairs_hist_gradient_boosting(
                attrs, False, tree, tree_id, tree_weight, 0, False, False, dtype=dtype
            )

    if hasattr(op, "_baseline_prediction"):
        if isinstance(op._baseline_prediction, np.ndarray):
            attrs["base_values"] = list(op._baseline_prediction)
        else:
            attrs["base_values"] = [op._baseline_prediction]

    input_name = operator.input_full_names
    if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
        cast_input_name = scope.get_unique_variable_name("cast_input")

        apply_cast(
            scope,
            operator.input_full_names,
            cast_input_name,
            container,
            to=onnx_proto.TensorProto.FLOAT,
        )
        input_name = cast_input_name

    if dtype is not None:
        for k in attrs:
            if k in (
                "nodes_values",
                "class_weights",
                "target_weights",
                "nodes_hitrates",
                "base_values",
            ):
                attrs[k] = np.array(attrs[k], dtype=dtype).ravel()

    container.add_node(
        op_type,
        input_name,
        operator.outputs[0].full_name,
        op_domain=op_domain,
        op_version=op_version,
        **attrs
    )

    if hasattr(op, "n_trees_per_iteration_"):
        # HistGradientBoostingRegressor does not implement decision_path.
        return
    if isinstance(op, RandomTreesEmbedding):
        options = scope.get_options(op)
    else:
        options = scope.get_options(op, dict(decision_path=False, decision_leaf=False))

    if not options.get("decision_path", False) and not options.get(
        "decision_leaf", False
    ):
        return

    # decision_path
    tree_paths = []
    tree_leaves = []
    for i, tree in enumerate(op.estimators_):
        attrs = get_default_tree_regressor_attribute_pairs()
        attrs["name"] = scope.get_unique_operator_name("%s_%d" % (op_type, i))
        attrs["n_targets"] = int(op.n_outputs_)
        add_tree_to_attribute_pairs(
            attrs, False, tree.tree_, 0, 1.0, 0, False, True, dtype=dtype
        )

        attrs["n_targets"] = 1
        attrs["post_transform"] = "NONE"
        attrs["target_ids"] = [0 for _ in attrs["target_ids"]]
        attrs["target_weights"] = [float(_) for _ in attrs["target_nodeids"]]

        if dtype is not None:
            for k in attrs:
                if k in (
                    "nodes_values",
                    "class_weights",
                    "target_weights",
                    "nodes_hitrates",
                    "base_values",
                ):
                    attrs[k] = np.array(attrs[k], dtype=dtype).ravel()

        if options.get("decision_path", False):
            # decision_path
            tree_paths.append(
                _append_decision_output(
                    input_name,
                    attrs,
                    _build_labels_path,
                    None,
                    scope,
                    operator,
                    container,
                    op_type=op_type,
                    op_domain=op_domain,
                    op_version=op_version,
                    regression=True,
                    overwrite_tree=tree.tree_,
                )
            )
        if options.get("decision_leaf", False):
            # decision_path
            tree_leaves.append(
                _append_decision_output(
                    input_name,
                    attrs,
                    _build_labels_leaf,
                    None,
                    scope,
                    operator,
                    container,
                    op_type=op_type,
                    op_domain=op_domain,
                    op_version=op_version,
                    regression=True,
                    cast_encode=True,
                )
            )

    # merges everything
    n_out = 1
    if options.get("decision_path", False):
        apply_concat(
            scope,
            tree_paths,
            operator.outputs[n_out].full_name,
            container,
            axis=1,
            operator_name=scope.get_unique_operator_name("concat"),
        )
        n_out += 1

    if options.get("decision_leaf", False):
        # decision_path
        apply_concat(
            scope,
            tree_leaves,
            operator.outputs[n_out].full_name,
            container,
            axis=1,
            operator_name=scope.get_unique_operator_name("concat"),
        )
        n_out += 1


register_converter(
    "SklearnRandomForestClassifier",
    convert_sklearn_random_forest_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "raw_scores": [True, False],
        "nocl": [True, False],
        "output_class_labels": [False, True],
        "decision_path": [True, False],
        "decision_leaf": [True, False],
    },
)
register_converter(
    "SklearnRandomForestRegressor",
    convert_sklearn_random_forest_regressor_converter,
    options={"decision_path": [True, False], "decision_leaf": [True, False]},
)
register_converter(
    "SklearnExtraTreesClassifier",
    convert_sklearn_random_forest_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "raw_scores": [True, False],
        "nocl": [True, False],
        "output_class_labels": [False, True],
        "decision_path": [True, False],
        "decision_leaf": [True, False],
    },
)
register_converter(
    "SklearnExtraTreesRegressor",
    convert_sklearn_random_forest_regressor_converter,
    options={"decision_path": [True, False], "decision_leaf": [True, False]},
)
register_converter(
    "SklearnHistGradientBoostingClassifier",
    convert_sklearn_random_forest_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "raw_scores": [True, False],
        "output_class_labels": [False, True],
        "nocl": [True, False],
    },
)
register_converter(
    "SklearnHistGradientBoostingRegressor",
    convert_sklearn_random_forest_regressor_converter,
    options={
        "zipmap": [True, False, "columns"],
        "raw_scores": [True, False],
        "output_class_labels": [False, True],
        "nocl": [True, False],
    },
)



# SPDX-License-Identifier: Apache-2.0
import numpy as np
from ..proto import onnx_proto
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_class_labels(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    if getattr(operator, "is_multi_output", False):
        classes = operator.classes
        if not isinstance(classes, list):
            raise RuntimeError(
                "classes must be a list of numpy arrays but is %r." "" % type(classes)
            )
        names = []
        if classes[0].dtype in (np.int64, np.int32):
            for i, cl in enumerate(classes):
                cla = np.array(cl)
                name = scope.get_unique_variable_name(
                    operator.outputs[0].full_name + "_cst_%d" % i
                )
                container.add_initializer(
                    name, onnx_proto.TensorProto.INT64, list(cla.shape), cla.tolist()
                )
                names.append(name)
        else:
            for i, cl in enumerate(classes):
                name = scope.get_unique_variable_name(
                    operator.outputs[0].full_name + "_cst_%d" % i
                )
                clids = np.arange(len(cl), dtype=np.int64)
                container.add_initializer(
                    name,
                    onnx_proto.TensorProto.INT64,
                    list(clids.shape),
                    clids.tolist(),
                )
                namele = scope.get_unique_variable_name(
                    operator.outputs[0].full_name + "_le_%d" % i
                )
                container.add_node(
                    "LabelEncoder",
                    name,
                    namele,
                    op_domain="ai.onnx.ml",
                    op_version=2,
                    default_string="0",
                    keys_int64s=clids,
                    values_strings=cl.tolist(),
                    name=scope.get_unique_operator_name("class_labels_le_%d" % i),
                )
                names.append(namele)

        container.add_node(
            "SequenceConstruct",
            names,
            operator.outputs[0].full_name,
            name=scope.get_unique_operator_name("class_labels_seq"),
        )
    else:
        classes = np.array(operator.classes)
        name = scope.get_unique_variable_name(operator.outputs[0].full_name + "_cst")

        if classes.dtype in (np.int64, np.int32):
            container.add_initializer(
                name,
                onnx_proto.TensorProto.INT64,
                list(classes.shape),
                classes.tolist(),
            )
        else:
            clids = np.arange(len(classes), dtype=np.int64)
            container.add_initializer(
                name, onnx_proto.TensorProto.INT64, list(clids.shape), clids.tolist()
            )
            namele = scope.get_unique_variable_name(
                operator.outputs[0].full_name + "_le"
            )
            container.add_node(
                "LabelEncoder",
                name,
                namele,
                op_domain="ai.onnx.ml",
                op_version=2,
                default_string="0",
                keys_int64s=clids,
                values_strings=classes.tolist(),
                name=scope.get_unique_operator_name("class_labels_le"),
            )
            name = namele

        container.add_node("Identity", name, operator.outputs[0].full_name)


register_converter("SklearnClassLabels", convert_sklearn_class_labels)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..proto import onnx_proto
from ..common._registration import register_converter
from ..common.data_types import Int64TensorType, StringTensorType
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._apply_operation import apply_concat
from .common import concatenate_variables


def convert_sklearn_imputer(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    op_type = "Imputer"
    attrs = {"name": scope.get_unique_operator_name(op_type)}
    op = operator.raw_operator
    if (
        hasattr(op, "fill_value")
        and isinstance(op.fill_value, str)
        and op.fill_value.lower() != "nan"
    ):
        raise RuntimeError(
            "Imputer cannot fill missing values with a " "string '%s'." % op.fill_value
        )
    if not hasattr(op, "statistics_"):
        raise RuntimeError(
            "Member statistics_ is not present, was the " "model fitted?"
        )

    if isinstance(operator.inputs[0].type, StringTensorType):
        if not isinstance(op.missing_values, (str, np.str_)):
            raise NotImplementedError(
                "The converter is implemented when the missing values "
                "are string not %r." % type(op.missing_values)
            )

        zero = scope.get_unique_variable_name("zero")
        container.add_initializer(zero, onnx_proto.TensorProto.INT64, [1], [0])

        concatenated_feature = concatenate_variables(scope, operator.inputs, container)
        names = []
        for i in range(op.statistics_.size):
            # loop on features
            fill_value = scope.get_unique_variable_name("fillvalue")
            if op.fill_value is None:
                skl_fill_value = op.statistics_[i]
            else:
                skl_fill_value = op.fill_value
            container.add_node(
                "LabelEncoder",
                [zero],
                [fill_value],
                keys_int64s=[0],
                values_strings=[op.statistics_[i]],
                default_string=skl_fill_value,
                op_domain="ai.onnx.ml",
                op_version=2,
            )
            init = scope.get_unique_variable_name("i%d" % i)
            container.add_initializer(init, onnx_proto.TensorProto.INT64, [1], [i])
            name = scope.get_unique_variable_name("impi%d" % i)
            container.add_node(
                "ArrayFeatureExtractor",
                [concatenated_feature, init],
                [name],
                op_domain="ai.onnx.ml",
            )
            cond = scope.get_unique_variable_name("impc%d" % i)
            container.add_node(
                "LabelEncoder",
                [name],
                [cond],
                keys_strings=[str(op.missing_values)],
                values_int64s=[1],
                default_int64=0,
                op_domain="ai.onnx.ml",
                op_version=2,
            )
            condb = scope.get_unique_variable_name("impc%d" % i)
            container.add_node("Cast", [cond], [condb], to=onnx_proto.TensorProto.BOOL)

            repli = scope.get_unique_variable_name("nomiss%d" % i)
            container.add_node("Where", [condb, fill_value, name], [repli])
            names.append(repli)

        apply_concat(scope, names, operator.outputs[0].full_name, container, axis=1)
    else:
        if isinstance(operator.inputs[0].type, Int64TensorType):
            attrs["imputed_value_int64s"] = op.statistics_.astype(np.int64)
            use_int = True
            delta = np.max(np.abs(attrs["imputed_value_int64s"] - op.statistics_))
            if delta != 0:
                raise RuntimeError(
                    "SimpleImputer takes integer as input but nan values are "
                    "replaced by float {} != {}.".format(
                        attrs["imputed_value_int64s"], op.statistics_
                    )
                )
        else:
            attrs["imputed_value_floats"] = op.statistics_.astype(np.float32)
            use_int = False

        if isinstance(op.missing_values, str) and op.missing_values == "NaN":
            attrs["replaced_value_float"] = np.NaN
        elif isinstance(op.missing_values, float):
            if use_int:
                ar = np.array([op.missing_values]).astype(np.int64)
                attrs["replaced_value_int64"] = ar[0]
            else:
                attrs["replaced_value_float"] = float(op.missing_values)
        else:
            raise RuntimeError(
                "Unsupported proposed value '{0}'. You may raise an issue at "
                "https://github.com/onnx/sklearn-onnx/issues."
                "".format(op.missing_values)
            )

        concatenated_feature = concatenate_variables(scope, operator.inputs, container)
        container.add_node(
            op_type,
            concatenated_feature,
            operator.outputs[0].full_name,
            op_domain="ai.onnx.ml",
            **attrs
        )


register_converter("SklearnImputer", convert_sklearn_imputer)
register_converter("SklearnSimpleImputer", convert_sklearn_imputer)



# SPDX-License-Identifier: Apache-2.0

from ..proto import onnx_proto
from ..common._apply_operation import (
    apply_slice,
    apply_cast,
    apply_identity,
    apply_reshape,
)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def _common_convert_sklearn_zipmap(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    zipmap_attrs = {"name": scope.get_unique_operator_name("ZipMap")}
    to_type = onnx_proto.TensorProto.INT64

    if hasattr(operator, "classlabels_int64s"):
        zipmap_attrs["classlabels_int64s"] = operator.classlabels_int64s
    elif hasattr(operator, "classlabels_strings"):
        zipmap_attrs["classlabels_strings"] = operator.classlabels_strings
        to_type = onnx_proto.TensorProto.STRING

    if to_type == onnx_proto.TensorProto.STRING:
        apply_identity(
            scope,
            operator.inputs[0].full_name,
            operator.outputs[0].full_name,
            container,
        )
    else:
        apply_cast(
            scope,
            operator.inputs[0].full_name,
            operator.outputs[0].full_name,
            container,
            to=to_type,
        )
    return zipmap_attrs


def convert_sklearn_zipmap(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    if len(operator.inputs) == 2:
        zipmap_attrs = _common_convert_sklearn_zipmap(scope, operator, container)
        container.add_node(
            "ZipMap",
            operator.inputs[1].full_name,
            operator.outputs[1].full_name,
            op_domain="ai.onnx.ml",
            **zipmap_attrs
        )
        return

    if hasattr(operator, "classlabels_int64s"):
        zipmap_attrs = dict(classlabels_int64s=operator.classlabels_int64s)
    elif hasattr(operator, "classlabels_strings"):
        zipmap_attrs = dict(classlabels_strings=operator.classlabels_strings)
    else:
        raise RuntimeError(
            "operator should have attribute 'classlabels_int64s' or "
            "'classlabels_strings'."
        )
    container.add_node(
        "ZipMap",
        operator.inputs[0].full_name,
        operator.outputs[0].full_name,
        op_domain="ai.onnx.ml",
        **zipmap_attrs
    )


def convert_sklearn_zipmap_columns(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    _common_convert_sklearn_zipmap(scope, operator, container)
    probs = operator.inputs[1].full_name
    for i in range(1, len(operator.outputs)):
        out = operator.outputs[i].full_name
        flat = scope.get_unique_variable_name(out)
        apply_slice(
            scope,
            probs,
            flat,
            container,
            starts=[i - 1],
            ends=[i],
            axes=[1],
            operator_name=scope.get_unique_operator_name("Slice"),
        )
        apply_reshape(
            scope,
            flat,
            out,
            container,
            desired_shape=(-1,),
            operator_name=scope.get_unique_operator_name("reshape"),
        )


register_converter("SklearnZipMap", convert_sklearn_zipmap)
register_converter("SklearnZipMapColumns", convert_sklearn_zipmap_columns)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from ..common.data_types import Int64TensorType, guess_numpy_type
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..algebra.onnx_ops import (
    OnnxAdd,
    OnnxCast,
    OnnxExp,
    OnnxIdentity,
    OnnxMatMul,
    OnnxReshape,
    OnnxSigmoid,
)


def convert_sklearn_gamma_regressor(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    X = operator.inputs[0]
    out = operator.outputs
    op = operator.raw_operator
    opv = container.target_opset
    dtype = guess_numpy_type(X.type)
    if dtype != np.float64:
        dtype = np.float32

    if isinstance(X.type, Int64TensorType):
        input_var = OnnxCast(X, to=np.float32, op_version=opv)
    else:
        input_var = X

    intercept = (
        op.intercept_.astype(dtype)
        if len(op.intercept_.shape) > 0
        else np.array([op.intercept_], dtype=dtype)
    )
    eta = OnnxAdd(
        OnnxMatMul(input_var, op.coef_.astype(dtype), op_version=opv),
        intercept,
        op_version=opv,
    )

    if hasattr(op, "_link_instance"):
        # scikit-learn < 1.1
        from sklearn.linear_model._glm.link import IdentityLink, LogLink, LogitLink

        if isinstance(op._link_instance, IdentityLink):
            Y = OnnxIdentity(eta, op_version=opv)
        elif isinstance(op._link_instance, LogLink):
            Y = OnnxExp(eta, op_version=opv)
        elif isinstance(op._link_instance, LogitLink):
            Y = OnnxSigmoid(eta, op_version=opv)
        else:
            raise RuntimeError(
                "Unexpected type %r for _link_instance "
                "in operator type %r." % (type(op._link_instance), type(op))
            )
    else:
        # scikit-learn >= 1.1
        from sklearn._loss.loss import (
            AbsoluteError,
            HalfBinomialLoss,
            HalfGammaLoss,
            HalfPoissonLoss,
            HalfSquaredError,
            HalfTweedieLoss,
            HalfTweedieLossIdentity,
            PinballLoss,
        )

        loss = op._get_loss()
        if isinstance(
            loss,
            (AbsoluteError, HalfSquaredError, HalfTweedieLossIdentity, PinballLoss),
        ):
            Y = OnnxIdentity(eta, op_version=opv)
        elif isinstance(loss, (HalfPoissonLoss, HalfGammaLoss, HalfTweedieLoss)):
            Y = OnnxExp(eta, op_version=opv)
        elif isinstance(loss, HalfBinomialLoss):
            Y = OnnxSigmoid(eta, op_version=opv)
        else:
            raise RuntimeError(
                f"Unexpected type of link for {loss!r} loss " "in operator type {op!r}."
            )

    last_dim = 1 if len(op.coef_.shape) == 1 else op.coef_.shape[-1]
    final = OnnxReshape(
        Y,
        np.array([-1, last_dim], dtype=np.int64),
        op_version=opv,
        output_names=out[:1],
    )
    final.add_to(scope, container)


register_converter("SklearnGammaRegressor", convert_sklearn_gamma_regressor)



# SPDX-License-Identifier: Apache-2.0
import copy

import numpy as np

from ..common._apply_operation import apply_cast, apply_concat, apply_reshape
from ..common._container import ModelComponentContainer
from ..common.data_types import (
    DoubleTensorType,
    FloatTensorType,
    Int64TensorType,
    Int32TensorType,
    Int16TensorType,
)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..proto import onnx_proto


def convert_sklearn_ordinal_encoder(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    ordinal_op = operator.raw_operator
    result = []
    input_idx = 0
    dimension_idx = 0
    for categories in ordinal_op.categories_:
        if len(categories) == 0:
            continue

        current_input = operator.inputs[input_idx]
        if current_input.get_second_dimension() == 1:
            feature_column = current_input
            input_idx += 1
        else:
            index_name = scope.get_unique_variable_name("index")
            container.add_initializer(
                index_name, onnx_proto.TensorProto.INT64, [], [dimension_idx]
            )

            feature_column = scope.declare_local_variable(
                "feature_column",
                current_input.type.__class__([current_input.get_first_dimension(), 1]),
            )

            container.add_node(
                "ArrayFeatureExtractor",
                [current_input.onnx_name, index_name],
                feature_column.onnx_name,
                op_domain="ai.onnx.ml",
                name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
            )

            dimension_idx += 1
            if dimension_idx == current_input.get_second_dimension():
                dimension_idx = 0
                input_idx += 1

        to = None
        if isinstance(current_input.type, DoubleTensorType):
            to = onnx_proto.TensorProto.FLOAT
        if isinstance(current_input.type, (Int16TensorType, Int32TensorType)):
            to = onnx_proto.TensorProto.INT64
        if to is not None:
            dtype = (
                Int64TensorType
                if to == onnx_proto.TensorProto.INT64
                else FloatTensorType
            )
            casted_feature_column = scope.declare_local_variable(
                "casted_feature_column", dtype(copy.copy(feature_column.type.shape))
            )

            apply_cast(
                scope,
                feature_column.onnx_name,
                casted_feature_column.onnx_name,
                container,
                to=to,
            )

            feature_column = casted_feature_column

        attrs = {"name": scope.get_unique_operator_name("LabelEncoder")}
        if isinstance(feature_column.type, FloatTensorType):
            attrs["keys_floats"] = np.array(
                [float(s) for s in categories], dtype=np.float32
            )
        elif isinstance(feature_column.type, Int64TensorType):
            attrs["keys_int64s"] = np.array(
                [int(s) for s in categories], dtype=np.int64
            )
        else:
            attrs["keys_strings"] = np.array(
                [str(s).encode("utf-8") for s in categories]
            )
        attrs["values_int64s"] = np.arange(len(categories)).astype(np.int64)

        result.append(scope.get_unique_variable_name("ordinal_output"))
        label_encoder_output = scope.get_unique_variable_name("label_encoder")

        container.add_node(
            "LabelEncoder",
            feature_column.onnx_name,
            label_encoder_output,
            op_domain="ai.onnx.ml",
            op_version=2,
            **attrs
        )
        apply_reshape(
            scope,
            label_encoder_output,
            result[-1],
            container,
            desired_shape=(-1, 1),
        )

    concat_result_name = scope.get_unique_variable_name("concat_result")
    apply_concat(scope, result, concat_result_name, container, axis=1)
    cast_type = (
        onnx_proto.TensorProto.FLOAT
        if np.issubdtype(ordinal_op.dtype, np.floating)
        else onnx_proto.TensorProto.INT64
    )
    apply_cast(
        scope, concat_result_name, operator.output_full_names, container, to=cast_type
    )


register_converter("SklearnOrdinalEncoder", convert_sklearn_ordinal_encoder)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.utils import check_input_and_output_numbers
from ..common.data_types import SequenceType

_stack = []


def multioutput_regressor_shape_calculator(operator):
    """Shape calculator for MultiOutputRegressor"""
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    i = operator.inputs[0]
    o = operator.outputs[0]
    N = i.get_first_dimension()
    C = len(operator.raw_operator.estimators_)
    o.type = o.type.__class__([N, C])


def multioutput_classifier_shape_calculator(operator):
    """Shape calculator for MultiOutputClassifier"""
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=2)
    if not isinstance(operator.outputs[1].type, SequenceType):
        raise RuntimeError(
            "Probabilites should be a sequence not %r." "" % operator.outputs[1].type
        )
    i = operator.inputs[0]
    outputs = operator.outputs
    N = i.get_first_dimension()
    C = len(operator.raw_operator.estimators_)
    outputs[0].type.shape = [N, C]


register_shape_calculator(
    "SklearnMultiOutputRegressor", multioutput_regressor_shape_calculator
)
register_shape_calculator(
    "SklearnMultiOutputClassifier", multioutput_classifier_shape_calculator
)



# SPDX-License-Identifier: Apache-2.0


import numbers
import numpy as np
from ..common._apply_operation import apply_cast
from ..common.data_types import BooleanTensorType, Int64TensorType, guess_numpy_type
from ..common._registration import register_converter
from ..common.tree_ensemble import (
    add_tree_to_attribute_pairs,
    get_default_tree_classifier_attribute_pairs,
    get_default_tree_regressor_attribute_pairs,
)
from ..proto import onnx_proto


def convert_sklearn_gradient_boosting_classifier(
    scope,
    operator,
    container,
    op_type="TreeEnsembleClassifier",
    op_domain="ai.onnx.ml",
    op_version=1,
):
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    op = operator.raw_operator
    if op.loss not in ("deviance", "log_loss"):
        raise NotImplementedError(
            "Loss '{0}' is not supported yet. You "
            "may raise an issue at "
            "https://github.com/onnx/sklearn-onnx/issues.".format(op.loss)
        )

    attrs = get_default_tree_classifier_attribute_pairs()
    attrs["name"] = scope.get_unique_operator_name(op_type)

    transform = "LOGISTIC" if op.n_classes_ == 2 else "SOFTMAX"
    if op.init == "zero":
        loss = op._loss if hasattr(op, "_loss") else op.loss_
        if hasattr(loss, "K"):
            base_values = np.zeros(loss.K)
        else:
            base_values = np.zeros(1)
    elif op.init is None:
        if hasattr(op.estimators_[0, 0], "n_features_in_"):
            # sklearn >= 1.2
            n_features = op.estimators_[0, 0].n_features_in_
        else:
            # sklearn < 1.2
            n_features = op.estimators_[0, 0].n_features_
        x0 = np.zeros((1, n_features))
        if hasattr(op, "_raw_predict_init"):
            # sklearn >= 0.21
            base_values = op._raw_predict_init(x0).ravel()
        elif hasattr(op, "_init_decision_function"):
            # sklearn >= 0.20 and sklearn < 0.21
            base_values = op._init_decision_function(x0).ravel()
        else:
            raise RuntimeError("scikit-learn < 0.19 is not supported.")
    else:
        raise NotImplementedError(
            "Setting init to an estimator is not supported, you may raise an "
            "issue at https://github.com/onnx/sklearn-onnx/issues."
        )

    attrs["base_values"] = [float(v) for v in base_values]
    options = container.get_options(op, dict(raw_scores=False))
    if not options["raw_scores"]:
        attrs["post_transform"] = transform

    classes = op.classes_
    if all(isinstance(i, (numbers.Real, bool, np.bool_)) for i in classes):
        class_labels = [int(i) for i in classes]
        attrs["classlabels_int64s"] = class_labels
    elif all(isinstance(i, str) for i in classes):
        class_labels = [str(i) for i in classes]
        attrs["classlabels_strings"] = class_labels
    else:
        raise ValueError("Labels must be all integer or all strings.")

    tree_weight = op.learning_rate
    n_est = op.n_estimators_ if hasattr(op, "n_estimators_") else op.n_estimators
    if op.n_classes_ == 2:
        for tree_id in range(n_est):
            tree = op.estimators_[tree_id][0].tree_
            add_tree_to_attribute_pairs(
                attrs, True, tree, tree_id, tree_weight, 0, False, True, dtype=dtype
            )
    else:
        for i in range(n_est):
            for c in range(op.n_classes_):
                tree_id = i * op.n_classes_ + c
                tree = op.estimators_[i][c].tree_
                add_tree_to_attribute_pairs(
                    attrs, True, tree, tree_id, tree_weight, c, False, True, dtype=dtype
                )

    if dtype is not None:
        for k in attrs:
            if k in (
                "nodes_values",
                "class_weights",
                "target_weights",
                "nodes_hitrates",
                "base_values",
            ):
                attrs[k] = np.array(attrs[k], dtype=dtype)

    input_name = operator.input_full_names
    if isinstance(operator.inputs[0].type, BooleanTensorType):
        cast_input_name = scope.get_unique_variable_name("cast_input")

        apply_cast(
            scope,
            input_name,
            cast_input_name,
            container,
            to=onnx_proto.TensorProto.FLOAT,
        )
        input_name = cast_input_name
    container.add_node(
        op_type,
        input_name,
        [operator.outputs[0].full_name, operator.outputs[1].full_name],
        op_domain=op_domain,
        op_version=op_version,
        **attrs
    )


def convert_sklearn_gradient_boosting_regressor(
    scope,
    operator,
    container,
    op_type="TreeEnsembleRegressor",
    op_domain="ai.onnx.ml",
    op_version=1,
):
    op = operator.raw_operator
    attrs = get_default_tree_regressor_attribute_pairs()
    attrs["name"] = scope.get_unique_operator_name(op_type)
    attrs["n_targets"] = 1

    if op.init == "zero":
        loss = op._loss if hasattr(op, "_loss") else op.loss_
        if hasattr(loss, "K"):
            cst = np.zeros(loss.K)
        else:
            cst = np.zeros(1)
    elif op.init is None:
        # constant_ was introduced in scikit-learn 0.21.
        if hasattr(op.init_, "constant_"):
            if len(op.init_.constant_.shape) == 0:
                cst = [float(op.init_.constant_)]
            else:
                cst = [float(x) for x in op.init_.constant_.ravel().tolist()]
        elif op.loss == "ls":
            cst = [op.init_.mean]
        else:
            cst = [op.init_.quantile]
    else:
        raise NotImplementedError(
            "Setting init to an estimator is not supported, you may raise an "
            "issue at https://github.com/onnx/sklearn-onnx/issues."
        )

    attrs["base_values"] = [float(x) for x in cst]

    tree_weight = op.learning_rate
    n_est = op.n_estimators_ if hasattr(op, "n_estimators_") else op.n_estimators
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    for i in range(n_est):
        tree = op.estimators_[i][0].tree_
        tree_id = i
        add_tree_to_attribute_pairs(
            attrs, False, tree, tree_id, tree_weight, 0, False, True, dtype=dtype
        )

    if dtype is not None:
        for k in attrs:
            if k in (
                "nodes_values",
                "class_weights",
                "target_weights",
                "nodes_hitrates",
                "base_values",
            ):
                attrs[k] = np.array(attrs[k], dtype=dtype)

    input_name = operator.input_full_names
    if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
        cast_input_name = scope.get_unique_variable_name("cast_input")

        apply_cast(
            scope,
            operator.input_full_names,
            cast_input_name,
            container,
            to=onnx_proto.TensorProto.FLOAT,
        )
        input_name = cast_input_name

    container.add_node(
        op_type,
        input_name,
        operator.output_full_names,
        op_domain=op_domain,
        op_version=op_version,
        **attrs
    )


register_converter(
    "SklearnGradientBoostingClassifier",
    convert_sklearn_gradient_boosting_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "raw_scores": [True, False],
        "output_class_labels": [False, True],
        "nocl": [True, False],
    },
)
register_converter(
    "SklearnGradientBoostingRegressor", convert_sklearn_gradient_boosting_regressor
)



# SPDX-License-Identifier: Apache-2.0

"""
Place holder for all ONNX operators.
"""
import sys
import os
import numpy as np

try:
    from scipy.sparse import coo_matrix
except ImportError:
    from scipy.sparse.coo import coo_matrix
import onnx
from ..common.data_types import DataType
from ..common._topology import Variable
from .automation import get_rst_doc
from ._cache import cache_folder


def ClassFactory(
    class_name,
    op_name,
    inputs,
    outputs,
    input_range,
    output_range,
    domain,
    attr_names,
    doc,
    deprecated,
    since_version,
    past_version,
):
    from .onnx_operator import OnnxOperator, OnnxOperatorItem

    def __init__(self, *args, **kwargs):
        op_version = kwargs.pop("op_version", None)
        if isinstance(op_version, dict):
            op_version = op_version.get(domain, None)

        if op_version is None:
            if len(args) == 0 and input_range[0] == input_range[1]:
                args = [_[0] for _ in self.__class__.expected_inputs]
            if not (input_range[0] <= len(args) <= input_range[1]):
                raise RuntimeError(
                    "Unexpected number of inputs, "
                    "got {}, expecting {} for operator "
                    "'{}'.".format(len(args), len(inputs), op_name)
                )

        attr_names = self.attr_names
        if "_" in self.__class__.__name__:
            op_version_class = int(self.__class__.__name__.split("_")[-1])
            if op_version is None:
                op_version = op_version_class
            try:
                op_version = min(op_version, op_version_class)
            except TypeError:
                raise TypeError(
                    "Could not compare versions {} ? {} for "
                    "class '{}' since_version {}. Parameter 'op_version' "
                    "is probably missing when the class "
                    "is instantiated.".format(
                        op_version, op_version_class, class_name, since_version
                    )
                )
        else:
            op_version_class = None

        # By default, the op_version is None.
        # None means the latest available.
        if op_version is None:
            op_version = since_version

        found = None
        if op_version is not None:
            # attr_names refers to the most recent version of
            # this operator. We may need an older one.
            for op in range(op_version, 0, -1):
                name = "{}_{}".format(self.__class__.__name__, op)
                if name in self.past_version:
                    found = (name, op)
                    attr_names = self.past_version[name].attr_names
                    break
        if (
            op_version_class is not None
            and found is not None
            and found[-1] != op_version_class
        ):
            raise RuntimeError(
                "op_version={} does not refer to the same opset as the class "
                "name ('{}').".format(op_version, self.__class__.__name__)
            )
        for key in kwargs:
            if key in {
                "output_names",
                "op_version",
                "domain",
                "ir_version",
                "global_context",
                "clear_subgraph_inputs",
            }:
                continue
            if key not in attr_names:
                raise TypeError(
                    "Argument '%s' not valid for '%s' opset=%s."
                    % (key, op_name, op_version)
                )

        if op_version is not None:
            kwargs["op_version"] = op_version
        # This class can only be created by a user. Let's check
        # types are either a variable, an operator or an array.
        for i, a in enumerate(args):
            if isinstance(a, tuple):
                if len(a) != 2:
                    raise TypeError(
                        "Input %r is a tuple or class %r, it must have two "
                        "elements (name, type) not %r." % (i, class_name, a)
                    )
                if not isinstance(a[0], str) or not isinstance(a[1], DataType):
                    raise TypeError(
                        "Input %r is a tuple or class %r, it must be a tuple "
                        "(name, type) not %r." % (i, class_name, a)
                    )
                continue
            if not isinstance(
                a,
                (Variable, OnnxOperator, np.ndarray, str, OnnxOperatorItem, coo_matrix),
            ):
                raise TypeError(
                    "Unexpected type %r for input %r of operator %r. "
                    "It must be an instance of Variable (or a string), "
                    "OnnxOperator, OnnxOperatorItem, numpy.ndarray, "
                    "coo_matrix)." % (type(a), i, class_name)
                )
        OnnxOperator.__init__(self, *args, **kwargs)

    newclass = type(
        class_name,
        (OnnxOperator,),
        {
            "__init__": __init__,
            "__doc__": doc,
            "expected_inputs": inputs,
            "expected_outputs": outputs,
            "operator_name": op_name,
            "input_range": input_range,
            "output_range": output_range,
            "domain": domain,
            "is_deprecated": deprecated,
            "since_version": since_version,
            "past_version": past_version,
            "attr_names": attr_names,
            "__module__": __name__,
        },
    )
    return newclass


def dynamic_class_creation(cache=False):
    """
    Automatically generates classes for each of the operators
    module *onnx* defines and described at
    `Operators
    <https://github.com/onnx/onnx/blob/master/docs/Operators.md>`_
    and `Operators
    <https://github.com/onnx/onnx/blob/master/docs/
    Operators-ml.md>`_.
    """
    cache_dir = cache_folder()
    res = {}
    for schema in onnx.defs.get_all_schemas_with_history():
        if schema.support_level == schema.SupportType.EXPERIMENTAL:
            # Skips experimental operators.
            continue
        # Multiple version can coexist. The last one is kept.
        if schema.name in res:
            if schema.since_version > res[schema.name].since_version:
                # We keep the most recent one.
                res[schema.name] = schema
        else:
            res[schema.name] = schema
        res[schema.name + "_" + str(schema.since_version)] = schema
    cls = {}

    def _c(obj, label, i):
        name = "%s%d" % (obj.name or label, i)
        try:
            tys = obj.type_str or ""
        except AttributeError:
            tys = obj.typeStr or ""
        return (name, tys)

    for name in sorted(res):
        schema = res[name]
        inputs = [_c(o, "I", i) for i, o in enumerate(schema.inputs)]
        outputs = [_c(o, "O", i) for i, o in enumerate(schema.outputs)]
        args = [p for p in schema.attributes]

        if "_" in name:
            class_name = "Onnx" + name
        else:
            class_name = "Onnx" + schema.name

        filename = os.path.join(
            cache_dir, schema.name + "_" + str(schema.since_version) + ".rst"
        )
        if not cache and os.path.exists(filename):
            with open(filename, "r", encoding="utf-8") as f:
                doc = f.read()
        else:
            doc = get_rst_doc(schema)
            if cache:
                with open(filename, "w", encoding="utf-8") as f:
                    f.write(doc)

        cl = ClassFactory(
            class_name,
            schema.name,
            inputs,
            outputs,
            [schema.min_input, schema.max_input],
            [schema.min_output, schema.max_output],
            schema.domain,
            args,
            "**Version**" + doc.split("**Version**")[-1],
            getattr(schema, "deprecated", False),
            schema.since_version,
            {},
        )
        cls[class_name] = cl

    # Retrieves past classes.
    for name in cls:
        if "_" not in name:
            continue
        main, version = name.split("_")
        last = cls[main]
        last.past_version[name] = cls[name]

    return cls


def _update_module():
    """
    Dynamically updates the module with operators defined
    by *ONNX*.
    """
    res = dynamic_class_creation()
    this = sys.modules[__name__]
    for k, v in res.items():
        setattr(this, k, v)


_update_module()


def OnnxReduceSumApi11(*x, axes=None, keepdims=1, op_version=None, output_names=None):
    """
    Adds operator ReduceSum with opset>=13 following API from opset 12.
    """
    if op_version is None:
        raise RuntimeError("op_version must be specified.")
    if op_version is None or op_version >= 13:
        if axes is None:
            return OnnxReduceSum(  # noqa
                *x, keepdims=keepdims, op_version=op_version, output_names=output_names
            )
        return OnnxReduceSum(  # noqa
            *x,
            np.array(axes, dtype=np.int64),
            keepdims=keepdims,
            op_version=op_version,
            output_names=output_names,
        )
    if op_version >= 11:
        if axes is None:
            return OnnxReduceSum_11(  # noqa
                *x, keepdims=keepdims, op_version=op_version, output_names=output_names
            )
        return OnnxReduceSum_11(  # noqa
            *x,
            axes=axes,
            keepdims=keepdims,
            op_version=op_version,
            output_names=output_names,
        )
    if axes is None:
        return OnnxReduceSum_1(
            *x,
            keepdims=keepdims,  # noqa
            op_version=op_version,
            output_names=output_names,
        )
    return OnnxReduceSum_1(
        *x,
        axes=axes,
        keepdims=keepdims,  # noqa
        op_version=op_version,
        output_names=output_names,
    )


def OnnxReduceAnyApi18(
    cl18, cl13, cl11, cl1, *x, axes=None, keepdims=1, op_version=None, output_names=None
):
    """
    Adds operator Reduce* with opset>=18 following API from opset 17.
    """
    if op_version is None:
        raise RuntimeError("op_version must be specified.")
    if op_version is None or op_version >= 18:
        if axes is None:
            return cl18(  # noqa
                *x, keepdims=keepdims, op_version=op_version, output_names=output_names
            )
        return cl18(  # noqa
            *x,
            np.array(axes, dtype=np.int64),
            keepdims=keepdims,
            op_version=op_version,
            output_names=output_names,
        )
    if op_version >= 13:
        if axes is None:
            return cl13(
                *x,
                keepdims=keepdims,  # noqa
                op_version=op_version,
                output_names=output_names,
            )
        return cl13(
            *x,
            axes=axes,
            keepdims=keepdims,  # noqa
            op_version=op_version,
            output_names=output_names,
        )
    if op_version >= 11:
        if axes is None:
            return cl11(
                *x,
                keepdims=keepdims,  # noqa
                op_version=op_version,
                output_names=output_names,
            )
        return cl11(
            *x,
            axes=axes,
            keepdims=keepdims,  # noqa
            op_version=op_version,
            output_names=output_names,
        )
    if axes is None:
        return cl1(
            *x,
            keepdims=keepdims,  # noqa
            op_version=op_version,
            output_names=output_names,
        )
    return cl1(
        *x,
        axes=axes,
        keepdims=keepdims,  # noqa
        op_version=op_version,
        output_names=output_names,
    )


def OnnxReduceSumSquareApi18(
    *x, axes=None, keepdims=1, op_version=None, output_names=None
):
    """
    Adds operator ReduceSumSquare with opset>=18 following API from opset 17.
    """
    if axes is None or not isinstance(axes, (list, np.ndarray)):
        raise TypeError(f"axes must be a list or an array not {type(axes)}.")
    return OnnxReduceAnyApi18(
        OnnxReduceSumSquare,
        OnnxReduceSumSquare_13,  # noqa
        OnnxReduceSumSquare_11,
        OnnxReduceSumSquare_1,  # noqa
        *x,
        axes=axes,
        keepdims=keepdims,
        op_version=op_version,
        output_names=output_names,
    )


def OnnxReduceMeanApi18(*x, axes=None, keepdims=1, op_version=None, output_names=None):
    """
    Adds operator ReduceMean with opset>=18 following API from opset 17.
    """
    return OnnxReduceAnyApi18(
        OnnxReduceMean,
        OnnxReduceMean_13,  # noqa
        OnnxReduceMean_11,
        OnnxReduceMean_1,  # noqa
        *x,
        axes=axes,
        keepdims=keepdims,
        op_version=op_version,
        output_names=output_names,
    )


def OnnxReduceMaxApi18(*x, axes=None, keepdims=1, op_version=None, output_names=None):
    """
    Adds operator ReduceMean with opset>=18 following API from opset 17.
    """
    return OnnxReduceAnyApi18(
        OnnxReduceMax,
        OnnxReduceMax_13,  # noqa
        OnnxReduceMax_11,
        OnnxReduceMax_1,  # noqa
        *x,
        axes=axes,
        keepdims=keepdims,
        op_version=op_version,
        output_names=output_names,
    )


def OnnxReduceLogSumExpApi18(
    *x, axes=None, keepdims=1, op_version=None, output_names=None
):
    """
    Adds operator ReduceMean with opset>=18 following API from opset 17.
    """
    return OnnxReduceAnyApi18(
        OnnxReduceLogSumExp,
        OnnxReduceLogSumExp_13,  # noqa
        OnnxReduceLogSumExp_11,
        OnnxReduceLogSumExp_1,  # noqa
        *x,
        axes=axes,
        keepdims=keepdims,
        op_version=op_version,
        output_names=output_names,
    )


def OnnxReduceL2Api18(*x, axes=None, keepdims=1, op_version=None, output_names=None):
    """
    Adds operator ReduceMean with opset>=18 following API from opset 17.
    """
    return OnnxReduceAnyApi18(
        OnnxReduceL2,
        OnnxReduceL2_13,  # noqa
        OnnxReduceL2_11,
        OnnxReduceL2_1,  # noqa
        *x,
        axes=axes,
        keepdims=keepdims,
        op_version=op_version,
        output_names=output_names,
    )


def OnnxSplitApi18(
    *x, axis=0, split=None, num_outputs=None, op_version=None, output_names=None
):
    """
    Adds operator Split with opset>=13 following API from opset 11.
    """
    if op_version is None:
        raise RuntimeError("op_version must be specified.")
    if op_version is None or op_version >= 18:
        if split is None:
            if num_outputs is None:
                if output_names is None:
                    raise RuntimeError(
                        "split or num_outputs or output_names "
                        "must be specified since opset 18."
                    )
                num_outputs = len(output_names)
            if num_outputs is None:
                raise AttributeError("num_outputs cannot be None for Split-18.")
            return OnnxSplit_18(  # noqa
                *x,
                axis=axis,
                op_version=op_version,
                num_outputs=num_outputs,
                output_names=output_names,
            )
        if num_outputs is None:
            return OnnxSplit_18(  # noqa
                *x,
                np.array(split, dtype=np.int64),
                axis=axis,
                op_version=op_version,
                output_names=output_names,
            )
        return OnnxSplit_18(  # noqa
            *x,
            np.array(split, dtype=np.int64),
            axis=axis,
            num_outputs=num_outputs,
            op_version=op_version,
            output_names=output_names,
        )
    if op_version >= 13:
        if split is None:
            return OnnxSplit_13(  # noqa
                *x, axis=axis, op_version=op_version, output_names=output_names
            )
        return OnnxSplit_13(  # noqa
            *x,
            np.array(split, dtype=np.int64),
            axis=axis,
            op_version=op_version,
            output_names=output_names,
        )
    if op_version >= 11:
        if split is None:
            return OnnxSplit_11(  # noqa
                *x, axis=axis, op_version=op_version, output_names=output_names
            )
        return OnnxSplit_11(  # noqa
            *x, split=split, axis=axis, op_version=op_version, output_names=output_names
        )
    if split is None:
        return OnnxSplit_2(  # noqa
            *x, axis=axis, op_version=op_version, output_names=output_names
        )
    return OnnxSplit_2(
        *x,
        split=split,
        axis=axis,  # noqa
        op_version=op_version,
        output_names=output_names,
    )


def OnnxSqueezeApi11(*x, axes=None, op_version=None, output_names=None):
    """
    Adds operator Squeeze with opset>=13 following API from opset 11.
    """
    if op_version is None:
        raise RuntimeError("op_version must be specified.")
    if op_version is None or op_version >= 13:
        return OnnxSqueeze(  # noqa
            *x,
            np.array(axes, dtype=np.int64),
            op_version=op_version,
            output_names=output_names,
        )
    if op_version >= 11:
        return OnnxSqueeze_11(  # noqa
            *x, axes=axes, op_version=op_version, output_names=output_names
        )
    return OnnxSqueeze_1(
        *x, axes=axes, op_version=op_version, output_names=output_names  # noqa
    )


def OnnxUnsqueezeApi11(*x, axes=None, op_version=None, output_names=None):
    """
    Adds operator Unsqueeze with opset>=13 following API from opset 11.
    """
    if op_version is None:
        raise RuntimeError("op_version must be specified.")
    if op_version is None or op_version >= 13:
        return OnnxUnsqueeze(  # noqa
            *x,
            np.array(axes, dtype=np.int64),
            op_version=op_version,
            output_names=output_names,
        )
    if op_version >= 11:
        return OnnxUnsqueeze_11(  # noqa
            *x, axes=axes, op_version=op_version, output_names=output_names
        )
    return OnnxUnsqueeze_1(
        *x, axes=axes, op_version=op_version, output_names=output_names  # noqa
    )


def OnnxReduceL2_typed(
    dtype, x, axes=None, keepdims=1, op_version=None, output_names=None
):
    """
    Adds operator ReduceL2 for float or double.
    """
    if dtype == np.float32:
        return OnnxReduceL2Api18(  # noqa
            x,
            axes=axes,
            keepdims=keepdims,
            op_version=op_version,
            output_names=output_names,
        )
    x2 = OnnxMul(x, x, op_version=op_version)  # noqa
    red = OnnxReduceSumApi11(x2, axes=[1], keepdims=1, op_version=op_version)
    return OnnxSqrt(red, op_version=op_version, output_names=output_names)  # noqa


def OnnxReshapeApi13(*x, allowzero=0, op_version=None, output_names=None):
    """
    Adds operator Reshape with opset>=14 following API from opset 13.
    """
    if op_version is None:
        raise RuntimeError("op_version must be specified.")
    if op_version is None or op_version >= 14:
        return OnnxReshape(  # noqa
            *x, allowzero=allowzero, op_version=op_version, output_names=output_names
        )
    if op_version >= 13:
        return OnnxReshape_13(  # noqa
            *x, op_version=op_version, output_names=output_names
        )
    return OnnxReshape_5(*x, op_version=op_version, output_names=output_names)  # noqa



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.utils import check_input_and_output_numbers


def calculate_sklearn_array_feature_extractor(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    i = operator.inputs[0]
    N = i.get_first_dimension()
    C = len(operator.column_indices)
    operator.outputs[0].type = i.type.__class__([N, C])


register_shape_calculator(
    "SklearnArrayFeatureExtractor", calculate_sklearn_array_feature_extractor
)



# SPDX-License-Identifier: Apache-2.0


from ..common._apply_operation import apply_cast, apply_sub
from ..common.data_types import BooleanTensorType, Int64TensorType, guess_proto_type
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..proto import onnx_proto


def convert_sklearn_sgd_oneclass_svm(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    input_name = operator.inputs[0].full_name
    output_names = operator.output_full_names
    model = operator.raw_operator
    coef = model.coef_.T
    offset = model.offset_

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    if isinstance(operator.inputs[0].type, (BooleanTensorType, Int64TensorType)):
        cast_input_name = scope.get_unique_variable_name("cast_input")
        apply_cast(
            scope, operator.input_full_names, cast_input_name, container, to=proto_dtype
        )
        input_name = cast_input_name

    coef_name = scope.get_unique_variable_name("coef")
    container.add_initializer(coef_name, proto_dtype, coef.shape, coef.ravel())

    offset_name = scope.get_unique_variable_name("offset")
    container.add_initializer(offset_name, proto_dtype, offset.shape, offset)

    matmul_result_name = scope.get_unique_variable_name("matmul_result")
    container.add_node(
        "MatMul",
        [input_name, coef_name],
        matmul_result_name,
        name=scope.get_unique_operator_name("MatMul"),
    )

    apply_sub(
        scope,
        [matmul_result_name, offset_name],
        output_names[1],
        container,
        broadcast=0,
    )

    pred = scope.get_unique_variable_name("class_prediction")
    container.add_node("Sign", output_names[1], pred, op_version=9)
    apply_cast(scope, pred, output_names[0], container, to=onnx_proto.TensorProto.INT64)


register_converter("SklearnSGDOneClassSVM", convert_sklearn_sgd_oneclass_svm)



# SPDX-License-Identifier: Apache-2.0

from .onnx_operator_mixin import OnnxOperatorMixin


class OnnxSubGraphOperatorMixin(OnnxOperatorMixin):
    """
    :class:`OnnxOperatorMixin` for converters.
    """

    pass



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.utils import check_input_and_output_numbers
from ..common.shape_calculator import _infer_linear_regressor_output_types


def voting_regressor_shape_calculator(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    return _infer_linear_regressor_output_types(operator)


register_shape_calculator("SklearnVotingRegressor", voting_regressor_shape_calculator)



# SPDX-License-Identifier: Apache-2.0


import numpy as np

from ..proto import onnx_proto
from ..common._apply_operation import (
    apply_cast,
    apply_concat,
    apply_reshape,
    apply_mul,
    apply_add,
)
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_k_bins_discretiser(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    op = operator.raw_operator

    if op.encode == "onehot":
        raise RuntimeError(
            "onehot encoding not supported. "
            "ONNX does not support sparse tensors. "
            "with opset < 11. You may raise an isue at "
            "https://github.com/onnx/sklearn-onnx/issues."
        )

    ranges = list(
        map(
            lambda e: e[1:-1] if len(e) > 2 else [np.finfo(np.float32).max],
            op.bin_edges_,
        )
    )
    digitised_output_name = [None] * len(ranges)
    last_column_name = None

    for i, item in enumerate(ranges):
        digitised_output_name[i] = scope.get_unique_variable_name(
            "digitised_output_{}".format(i)
        )
        column_index_name = scope.get_unique_variable_name("column_index")
        range_column_name = scope.get_unique_variable_name("range_column")
        column_name = scope.get_unique_variable_name("column")
        cast_column_name = scope.get_unique_variable_name("cast_column")
        less_result_name = scope.get_unique_variable_name("less_result")
        cast_result_name = scope.get_unique_variable_name("cast_result")
        concatenated_array_name = scope.get_unique_variable_name("concatenated_array")
        argmax_output_name = scope.get_unique_variable_name("argmax_output")

        container.add_initializer(
            column_index_name, onnx_proto.TensorProto.INT64, [], [i]
        )
        container.add_initializer(
            range_column_name, onnx_proto.TensorProto.FLOAT, [len(item)], item
        )

        container.add_node(
            "ArrayFeatureExtractor",
            [operator.inputs[0].full_name, column_index_name],
            column_name,
            name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
            op_domain="ai.onnx.ml",
        )
        apply_cast(
            scope,
            column_name,
            cast_column_name,
            container,
            to=onnx_proto.TensorProto.FLOAT,
        )
        container.add_node(
            "Less",
            [cast_column_name, range_column_name],
            less_result_name,
            name=scope.get_unique_operator_name("Less"),
        )
        apply_cast(
            scope,
            less_result_name,
            cast_result_name,
            container,
            to=onnx_proto.TensorProto.FLOAT,
        )

        if last_column_name is None:
            last_column_name = scope.get_unique_variable_name("last_column")
            zero_float = scope.get_unique_variable_name("zero_float")
            one_float = scope.get_unique_variable_name("one_float")
            zero_column = scope.get_unique_variable_name("zero_column")
            container.add_initializer(
                one_float, onnx_proto.TensorProto.FLOAT, [1], np.ones(1)
            )
            container.add_initializer(
                zero_float, onnx_proto.TensorProto.FLOAT, [1], np.zeros(1)
            )
            apply_mul(
                scope,
                [cast_column_name, zero_float],
                zero_column,
                container,
                broadcast=1,
            )
            apply_add(
                scope,
                [zero_column, one_float],
                last_column_name,
                container,
                broadcast=1,
            )

        apply_concat(
            scope,
            [cast_result_name, last_column_name],
            concatenated_array_name,
            container,
            axis=1,
        )
        container.add_node(
            "ArgMax",
            concatenated_array_name,
            argmax_output_name,
            axis=1,
            name=scope.get_unique_operator_name("ArgMax"),
        )
        if op.encode == "onehot-dense":
            onehot_result_name = scope.get_unique_variable_name("onehot_result")

            container.add_node(
                "OneHotEncoder",
                argmax_output_name,
                onehot_result_name,
                name=scope.get_unique_operator_name("OneHotEncoder"),
                cats_int64s=list(range(op.n_bins_[i])),
                op_domain="ai.onnx.ml",
            )
            apply_reshape(
                scope,
                onehot_result_name,
                digitised_output_name[i],
                container,
                desired_shape=(-1, op.n_bins_[i]),
            )
        else:
            apply_cast(
                scope,
                argmax_output_name,
                digitised_output_name[i],
                container,
                to=onnx_proto.TensorProto.FLOAT,
            )
    apply_concat(
        scope, digitised_output_name, operator.outputs[0].full_name, container, axis=1
    )


register_converter("SklearnKBinsDiscretizer", convert_sklearn_k_bins_discretiser)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from sklearn.base import TransformerMixin, BaseEstimator

try:
    from sklearn.utils.validation import _deprecate_positional_args
except ImportError:

    def _deprecate_positional_args(x):
        return x  # noqa


class ReplaceTransformer(TransformerMixin, BaseEstimator):

    """
    Replaces a value by another one.
    It can be used to replace 0 by nan.

    Parameters
    ----------
    from_value : value to replace
    to_value : new value
    dtype: dtype of replaced values
    """  # noqa

    @_deprecate_positional_args
    def __init__(self, *, from_value=0, to_value=np.nan, dtype=np.float32):
        BaseEstimator.__init__(self)
        self.dtype = dtype
        self.from_value = from_value
        self.to_value = to_value

    def _replace(self, a):
        if hasattr(a, "todense"):
            if np.isnan(self.to_value) and self.from_value == 0:
                # implicit
                return a
            raise RuntimeError(
                "Unable to replace 0 by nan one value by another " "in sparse matrix."
            )
        return np.where(a == self.from_value, self.to_value, a)

    def fit(self, X, y=None, sample_weight=None):
        """
        Does nothing except checking *dtype* may be applied.
        """
        self._replace(X)
        return self

    def transform(self, X, y=None):
        """
        Casts array X.
        """
        return self._replace(X)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from sklearn.preprocessing import MaxAbsScaler
from sklearn.preprocessing import RobustScaler, StandardScaler
from ..algebra.onnx_ops import OnnxSub, OnnxDiv, OnnxCast, OnnxMul, OnnxClip, OnnxAdd
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import guess_numpy_type, guess_proto_type
from ..proto import onnx_proto
from .common import concatenate_variables


def convert_sklearn_scaler(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    # If there are multiple input variables, we need to combine them as a
    # whole tensor. Integer(s) would be converted to float(s).
    # Options div use true division instead of Scaler operator
    # which replaces a division by a multiplication.
    # This leads to discrepencies in some cases.
    if len(operator.inputs) > 1:
        feature_name = concatenate_variables(scope, operator.inputs, container)
    else:
        feature_name = operator.inputs[0].full_name
    C = operator.outputs[0].get_second_dimension()

    op = operator.raw_operator
    op_type = "Scaler"
    attrs = {"name": scope.get_unique_operator_name(op_type)}

    if isinstance(op, StandardScaler):
        model_C = None
        if op.scale_ is not None:
            model_C = op.scale_.shape[0]
        if model_C is None and op.mean_ is not None:
            model_C = op.mean_.shape[0]
        if model_C is None and op.var_ is not None:
            model_C = op.var_.shape[0]
        if model_C is None:
            # Identity
            container.add_node("Identity", feature_name, operator.outputs[0].full_name)
            return
        if C is not None and C != model_C:
            raise RuntimeError(
                "Unable Mismatch between expected shape %r and model (., %r)"
                " in operator %r." % (operator.outputs[0].type.shape, model_C, operator)
            )
        C = model_C
        attrs["offset"] = (
            op.mean_ if op.with_mean else np.array([0.0] * C, dtype=np.float32)
        )
        attrs["scale"] = (
            1.0 / op.scale_ if op.with_std else np.array([1.0] * C, dtype=np.float32)
        )
        inv_scale = op.scale_ if op.with_std else None
    elif isinstance(op, RobustScaler):
        model_C = None
        if op.center_ is not None:
            model_C = op.center_.shape[0]
        if model_C is None and op.scale_ is not None:
            model_C = op.scale_.shape[0]
        if model_C is None:
            # Identity
            container.add_node("Identity", feature_name, operator.outputs[0].full_name)
            return
        if C is not None and C != model_C:
            raise RuntimeError(
                "Unable Mismatch between expected shape %r and model (., %r)"
                " in operator %r." % (operator.outputs[0].type.shape, model_C, operator)
            )
        C = model_C
        attrs["offset"] = (
            op.center_ if op.with_centering else np.array([0.0] * C, dtype=np.float32)
        )
        attrs["scale"] = (
            1.0 / op.scale_
            if op.with_scaling
            else np.array([1.0] * C, dtype=np.float32)
        )
        inv_scale = op.scale_ if op.with_scaling else None
    elif isinstance(op, MaxAbsScaler):
        model_C = None
        if op.max_abs_ is not None:
            model_C = op.max_abs_.shape[0]
        if model_C is None and op.scale_ is not None:
            model_C = op.scale_.shape[0]
        if model_C is None:
            # Identity
            container.add_node("Identity", feature_name, operator.outputs[0].full_name)
            return
        if C is not None and C != model_C:
            raise RuntimeError(
                "Unable Mismatch between expected shape %r and model (., %r)"
                " in operator %r." % (operator.outputs[0].type.shape, model_C, operator)
            )
        C = model_C
        attrs["scale"] = 1.0 / op.scale_
        attrs["offset"] = np.array([0.0] * C, dtype=np.float32)
        inv_scale = op.scale_
    else:
        raise ValueError(
            "Only scikit-learn StandardScaler and RobustScaler "
            "are supported but got %s. You may raise "
            "an issue at "
            "https://github.com/onnx/sklearn-onnx/issues."
            "" % type(op)
        )

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    for k in attrs:
        v = attrs[k]
        if isinstance(v, np.ndarray) and v.dtype != dtype:
            attrs[k] = v.astype(dtype)

    use_scaler_op = container.is_allowed({"Scaler"})
    if not use_scaler_op or dtype == np.float64:
        opv = container.target_opset
        if inv_scale is None:
            sub = OnnxSub(
                feature_name,
                attrs["offset"].astype(dtype),
                op_version=opv,
                output_names=[operator.outputs[0].full_name],
            )
            sub.add_to(scope, container)
        else:
            sub = OnnxSub(feature_name, attrs["offset"].astype(dtype), op_version=opv)
            div = OnnxDiv(
                sub,
                inv_scale.astype(dtype),
                op_version=opv,
                output_names=[operator.outputs[0].full_name],
            )
            div.add_to(scope, container)
        return

    if inv_scale is not None:
        options = container.get_options(op, dict(div="std"))
        div = options["div"]
        if div == "div":
            opv = container.target_opset
            sub = OnnxSub(feature_name, attrs["offset"].astype(dtype), op_version=opv)
            div = OnnxDiv(
                sub,
                inv_scale.astype(dtype),
                op_version=opv,
                output_names=[operator.outputs[0].full_name],
            )
            div.add_to(scope, container)
            return
        if div == "div_cast":
            opv = container.target_opset
            cast = OnnxCast(
                feature_name, to=onnx_proto.TensorProto.DOUBLE, op_version=opv
            )
            sub = OnnxSub(cast, attrs["offset"].astype(np.float64), op_version=opv)
            div = OnnxDiv(sub, inv_scale.astype(np.float64), op_version=opv)
            cast = OnnxCast(
                div,
                to=proto_dtype,
                op_version=opv,
                output_names=[operator.outputs[0].full_name],
            )
            cast.add_to(scope, container)
            return

    if attrs["offset"].size != attrs["scale"].size:
        # Scaler does not accept different size for offset and scale.
        size = max(attrs["offset"].size, attrs["scale"].size)
        ones = np.ones(size, dtype=attrs["offset"].dtype)
        attrs["offset"] = attrs["offset"] * ones
        attrs["scale"] = attrs["scale"] * ones

    container.add_node(
        op_type,
        feature_name,
        operator.outputs[0].full_name,
        op_domain="ai.onnx.ml",
        **attrs
    )


def convert_sklearn_min_max_scaler(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    # If there are multiple input variables, we need to combine them as a
    # whole tensor. Integer(s) would be converted to float(s).
    # Options div use true division instead of Scaler operator
    # which replaces a division by a multiplication.
    # This leads to discrepencies in some cases.
    if len(operator.inputs) > 1:
        feature_name = concatenate_variables(scope, operator.inputs, container)
    else:
        feature_name = operator.inputs[0].full_name

    op = operator.raw_operator
    opv = container.target_opset

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32

    # X *= self.scale_
    # X += self.min_
    # if self.clip:
    #     np.clip(X, self.feature_range[0], self.feature_range[1], out=X)
    casted = OnnxCast(feature_name, to=proto_dtype, op_version=opv)
    scaled = OnnxMul(casted, op.scale_.astype(dtype), op_version=opv)

    if getattr(op, "clip", False):
        # parameter clip was introduced in scikit-learn 0.24
        offset = OnnxAdd(scaled, op.min_.astype(dtype), op_version=opv)

        clipped = OnnxClip(
            offset,
            np.array(op.feature_range[0], dtype=dtype),
            np.array(op.feature_range[1], dtype=dtype),
            op_version=opv,
            output_names=[operator.outputs[0].full_name],
        )
        clipped.add_to(scope, container)
    else:
        offset = OnnxAdd(
            scaled,
            op.min_.astype(dtype),
            op_version=opv,
            output_names=[operator.outputs[0].full_name],
        )
        offset.add_to(scope, container)


register_converter(
    "SklearnRobustScaler",
    convert_sklearn_scaler,
    options={"div": ["std", "div", "div_cast"]},
)
register_converter(
    "SklearnScaler", convert_sklearn_scaler, options={"div": ["std", "div", "div_cast"]}
)
register_converter("SklearnMinMaxScaler", convert_sklearn_min_max_scaler)
register_converter(
    "SklearnMaxAbsScaler",
    convert_sklearn_scaler,
    options={"div": ["std", "div", "div_cast"]},
)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_converter
from ..common._apply_operation import apply_concat, apply_identity
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_function_transformer(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    op = operator.raw_operator
    if op.func is not None:
        raise RuntimeError(
            "FunctionTransformer is not supported unless the "
            "transform function is None (= identity). "
            "You may raise an issue at "
            "https://github.com/onnx/sklearn-onnx/issues."
        )
    if len(operator.inputs) == 1:
        apply_identity(
            scope,
            operator.inputs[0].full_name,
            operator.outputs[0].full_name,
            container,
        )
    else:
        apply_concat(
            scope,
            [i.full_name for i in operator.inputs],
            operator.outputs[0].full_name,
            container,
            axis=1,
        )


register_converter("SklearnFunctionTransformer", convert_sklearn_function_transformer)



# SPDX-License-Identifier: Apache-2.0


from onnx.helper import make_tensor
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._apply_operation import apply_mul
from ..common.utils_classifier import _finalize_converter_classes
from ..common.data_types import guess_proto_type, Int64TensorType
from .._supported_operators import sklearn_operator_name_map
from ..proto import onnx_proto


def convert_voting_classifier(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converts a *VotingClassifier* into *ONNX* format.

    *predict_proba* is not defined by *scikit-learn* when ``voting='hard'``.
    The converted model still defines a probability vector equal to the
    highest probability obtained for each class over all estimators.

    *scikit-learn* enables both modes, transformer and predictor
    for the voting classifier. *ONNX* does not make this
    distinction and always creates two outputs, labels
    and probabilities.
    """
    if scope.get_options(operator.raw_operator, dict(nocl=False))["nocl"]:
        raise RuntimeError(
            "Option 'nocl' is not implemented for operator '{}'.".format(
                operator.raw_operator.__class__.__name__
            )
        )
    proto_dtype = guess_proto_type(operator.outputs[1].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT
    op = operator.raw_operator
    n_classes = len(op.classes_)

    classes_ind_name = scope.get_unique_variable_name("classes_ind")
    container.add_initializer(
        classes_ind_name,
        onnx_proto.TensorProto.INT64,
        (1, n_classes),
        list(range(n_classes)),
    )

    probs_names = []
    one_name = None
    for i, estimator in enumerate(op.estimators_):
        if estimator is None:
            continue

        op_type = sklearn_operator_name_map[type(estimator)]

        this_operator = scope.declare_local_operator(op_type, estimator)
        this_operator.inputs = operator.inputs

        label_name = scope.declare_local_variable("label_%d" % i, Int64TensorType())
        prob_name = scope.declare_local_variable(
            "voting_proba_%d" % i, operator.outputs[1].type.__class__()
        )
        this_operator.outputs.append(label_name)
        this_operator.outputs.append(prob_name)

        if op.voting == "hard":
            if one_name is None:
                shape_name = scope.get_unique_variable_name("shape")
                container.add_node(
                    "Shape",
                    prob_name.onnx_name,
                    shape_name,
                    name=scope.get_unique_operator_name("Shape"),
                )
                zero_name = scope.get_unique_variable_name("zero")
                container.add_node(
                    "ConstantOfShape",
                    shape_name,
                    zero_name,
                    name=scope.get_unique_operator_name("CoSA"),
                    value=make_tensor("value", proto_dtype, (1,), [0.0]),
                    op_version=9,
                )
                one_name = scope.get_unique_variable_name("one")
                container.add_node(
                    "ConstantOfShape",
                    shape_name,
                    one_name,
                    name=scope.get_unique_operator_name("CoSB"),
                    value=make_tensor("value", proto_dtype, (1,), [1.0]),
                    op_version=9,
                )

            argmax_output_name = scope.get_unique_variable_name("argmax_output")
            container.add_node(
                "ArgMax",
                prob_name.onnx_name,
                argmax_output_name,
                name=scope.get_unique_operator_name("ArgMax"),
                axis=1,
            )

            equal_name = scope.get_unique_variable_name("equal")
            container.add_node(
                "Equal",
                [argmax_output_name, classes_ind_name],
                equal_name,
                name=scope.get_unique_operator_name("Equal"),
            )

            max_proba_name = scope.get_unique_variable_name("probsmax")
            container.add_node(
                "Where",
                [equal_name, one_name, zero_name],
                max_proba_name,
                name=scope.get_unique_operator_name("Where"),
            )
            prob_name = max_proba_name
        else:
            prob_name = prob_name.onnx_name

        if op.weights is not None:
            val = op.weights[i] / op.weights.sum()
        else:
            val = 1.0 / len(op.estimators_)

        weights_name = scope.get_unique_variable_name("w%d" % i)
        container.add_initializer(weights_name, proto_dtype, [1], [val])
        wprob_name = scope.get_unique_variable_name("wprob_name")
        apply_mul(scope, [prob_name, weights_name], wprob_name, container, broadcast=1)
        probs_names.append(wprob_name)

    if op.flatten_transform in (False, None):
        container.add_node(
            "Sum",
            probs_names,
            operator.outputs[1].full_name,
            name=scope.get_unique_operator_name("Sum"),
        )
    else:
        raise NotImplementedError(
            "flatten_transform==True is not implemented yet. "
            "You may raise an issue at "
            "https://github.com/onnx/sklearn-onnx/issues."
        )

    # labels
    label_name = scope.get_unique_variable_name("label_name")
    container.add_node(
        "ArgMax",
        operator.outputs[1].full_name,
        label_name,
        name=scope.get_unique_operator_name("ArgMax"),
        axis=1,
    )
    _finalize_converter_classes(
        scope,
        label_name,
        operator.outputs[0].full_name,
        container,
        op.classes_,
        proto_dtype,
    )


register_converter(
    "SklearnVotingClassifier",
    convert_voting_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "output_class_labels": [False, True],
        "nocl": [True, False],
    },
)



# SPDX-License-Identifier: Apache-2.0

"""
Informations related to the :epkg:`ImageNet` competition.
"""

class_names = {
    0: "tench, Tinca tinca",
    1: "goldfish, Carassius auratus",
    2: "great white shark, white shark, man-eater, man-eating shark, "
    "Carcharodon carcharias",
    3: "tiger shark, Galeocerdo cuvieri",
    4: "hammerhead, hammerhead shark",
    5: "electric ray, crampfish, numbfish, torpedo",
    6: "stingray",
    7: "cock",
    8: "hen",
    9: "ostrich, Struthio camelus",
    10: "brambling, Fringilla montifringilla",
    11: "goldfinch, Carduelis carduelis",
    12: "house finch, linnet, Carpodacus mexicanus",
    13: "junco, snowbird",
    14: "indigo bunting, indigo finch, indigo bird, Passerina cyanea",
    15: "robin, American robin, Turdus migratorius",
    16: "bulbul",
    17: "jay",
    18: "magpie",
    19: "chickadee",
    20: "water ouzel, dipper",
    21: "kite",
    22: "bald eagle, American eagle, Haliaeetus leucocephalus",
    23: "vulture",
    24: "great grey owl, great gray owl, Strix nebulosa",
    25: "European fire salamander, Salamandra salamandra",
    26: "common newt, Triturus vulgaris",
    27: "eft",
    28: "spotted salamander, Ambystoma maculatum",
    29: "axolotl, mud puppy, Ambystoma mexicanum",
    30: "bullfrog, Rana catesbeiana",
    31: "tree frog, tree-frog",
    32: "tailed frog, bell toad, ribbed toad, tailed toad, Ascaphus trui",
    33: "loggerhead, loggerhead turtle, Caretta caretta",
    34: "leatherback turtle, leatherback, leathery turtle, " "Dermochelys coriacea",
    35: "mud turtle",
    36: "terrapin",
    37: "box turtle, box tortoise",
    38: "banded gecko",
    39: "common iguana, iguana, Iguana iguana",
    40: "American chameleon, anole, Anolis carolinensis",
    41: "whiptail, whiptail lizard",
    42: "agama",
    43: "frilled lizard, Chlamydosaurus kingi",
    44: "alligator lizard",
    45: "Gila monster, Heloderma suspectum",
    46: "green lizard, Lacerta viridis",
    47: "African chameleon, Chamaeleo chamaeleon",
    48: "Komodo dragon, Komodo lizard, dragon lizard, giant lizard, "
    "Varanus komodoensis",
    49: "African crocodile, Nile crocodile, Crocodylus niloticus",
    50: "American alligator, Alligator mississipiensis",
    51: "triceratops",
    52: "thunder snake, worm snake, Carphophis amoenus",
    53: "ringneck snake, ring-necked snake, ring snake",
    54: "hognose snake, puff adder, sand viper",
    55: "green snake, grass snake",
    56: "king snake, kingsnake",
    57: "garter snake, grass snake",
    58: "water snake",
    59: "vine snake",
    60: "night snake, Hypsiglena torquata",
    61: "boa constrictor, Constrictor constrictor",
    62: "rock python, rock snake, Python sebae",
    63: "Indian cobra, Naja naja",
    64: "green mamba",
    65: "sea snake",
    66: "horned viper, cerastes, sand viper, horned asp, Cerastes cornutus",
    67: "diamondback, diamondback rattlesnake, Crotalus adamanteus",
    68: "sidewinder, horned rattlesnake, Crotalus cerastes",
    69: "trilobite",
    70: "harvestman, daddy longlegs, Phalangium opilio",
    71: "scorpion",
    72: "black and gold garden spider, Argiope aurantia",
    73: "barn spider, Araneus cavaticus",
    74: "garden spider, Aranea diademata",
    75: "black widow, Latrodectus mactans",
    76: "tarantula",
    77: "wolf spider, hunting spider",
    78: "tick",
    79: "centipede",
    80: "black grouse",
    81: "ptarmigan",
    82: "ruffed grouse, partridge, Bonasa umbellus",
    83: "prairie chicken, prairie grouse, prairie fowl",
    84: "peacock",
    85: "quail",
    86: "partridge",
    87: "African grey, African gray, Psittacus erithacus",
    88: "macaw",
    89: "sulphur-crested cockatoo, Kakatoe galerita, Cacatua galerita",
    90: "lorikeet",
    91: "coucal",
    92: "bee eater",
    93: "hornbill",
    94: "hummingbird",
    95: "jacamar",
    96: "toucan",
    97: "drake",
    98: "red-breasted merganser, Mergus serrator",
    99: "goose",
    100: "black swan, Cygnus atratus",
    101: "tusker",
    102: "echidna, spiny anteater, anteater",
    103: "platypus, duckbill, duckbilled platypus, duck-billed "
    "platypus, Ornithorhynchus anatinus",
    104: "wallaby, brush kangaroo",
    105: "koala, koala bear, kangaroo bear, native bear, " "Phascolarctos cinereus",
    106: "wombat",
    107: "jellyfish",
    108: "sea anemone, anemone",
    109: "brain coral",
    110: "flatworm, platyhelminth",
    111: "nematode, nematode worm, roundworm",
    112: "conch",
    113: "snail",
    114: "slug",
    115: "sea slug, nudibranch",
    116: "chiton, coat-of-mail shell, sea cradle, polyplacophore",
    117: "chambered nautilus, pearly nautilus, nautilus",
    118: "Dungeness crab, Cancer magister",
    119: "rock crab, Cancer irroratus",
    120: "fiddler crab",
    121: "king crab, Alaska crab, Alaskan king crab, Alaska king "
    "crab, Paralithodes camtschatica",
    122: "American lobster, Northern lobster, Maine lobster, " "Homarus americanus",
    123: "spiny lobster, langouste, rock lobster, crawfish, " "crayfish, sea crawfish",
    124: "crayfish, crawfish, crawdad, crawdaddy",
    125: "hermit crab",
    126: "isopod",
    127: "white stork, Ciconia ciconia",
    128: "black stork, Ciconia nigra",
    129: "spoonbill",
    130: "flamingo",
    131: "little blue heron, Egretta caerulea",
    132: "American egret, great white heron, Egretta albus",
    133: "bittern",
    134: "crane",
    135: "limpkin, Aramus pictus",
    136: "European gallinule, Porphyrio porphyrio",
    137: "American coot, marsh hen, mud hen, water hen, Fulica americana",
    138: "bustard",
    139: "ruddy turnstone, Arenaria interpres",
    140: "red-backed sandpiper, dunlin, Erolia alpina",
    141: "redshank, Tringa totanus",
    142: "dowitcher",
    143: "oystercatcher, oyster catcher",
    144: "pelican",
    145: "king penguin, Aptenodytes patagonica",
    146: "albatross, mollymawk",
    147: "grey whale, gray whale, devilfish, Eschrichtius "
    "gibbosus, Eschrichtius robustus",
    148: "killer whale, killer, orca, grampus, sea wolf, Orcinus orca",
    149: "dugong, Dugong dugon",
    150: "sea lion",
    151: "Chihuahua",
    152: "Japanese spaniel",
    153: "Maltese dog, Maltese terrier, Maltese",
    154: "Pekinese, Pekingese, Peke",
    155: "Shih-Tzu",
    156: "Blenheim spaniel",
    157: "papillon",
    158: "toy terrier",
    159: "Rhodesian ridgeback",
    160: "Afghan hound, Afghan",
    161: "basset, basset hound",
    162: "beagle",
    163: "bloodhound, sleuthhound",
    164: "bluetick",
    165: "black-and-tan coonhound",
    166: "Walker hound, Walker foxhound",
    167: "English foxhound",
    168: "redbone",
    169: "borzoi, Russian wolfhound",
    170: "Irish wolfhound",
    171: "Italian greyhound",
    172: "whippet",
    173: "Ibizan hound, Ibizan Podenco",
    174: "Norwegian elkhound, elkhound",
    175: "otterhound, otter hound",
    176: "Saluki, gazelle hound",
    177: "Scottish deerhound, deerhound",
    178: "Weimaraner",
    179: "Staffordshire bullterrier, Staffordshire bull terrier",
    180: "American Staffordshire terrier, Staffordshire terrier, "
    "American pit bull terrier, pit bull terrier",
    181: "Bedlington terrier",
    182: "Border terrier",
    183: "Kerry blue terrier",
    184: "Irish terrier",
    185: "Norfolk terrier",
    186: "Norwich terrier",
    187: "Yorkshire terrier",
    188: "wire-haired fox terrier",
    189: "Lakeland terrier",
    190: "Sealyham terrier, Sealyham",
    191: "Airedale, Airedale terrier",
    192: "cairn, cairn terrier",
    193: "Australian terrier",
    194: "Dandie Dinmont, Dandie Dinmont terrier",
    195: "Boston bull, Boston terrier",
    196: "miniature schnauzer",
    197: "giant schnauzer",
    198: "standard schnauzer",
    199: "Scotch terrier, Scottish terrier, Scottie",
    200: "Tibetan terrier, chrysanthemum dog",
    201: "silky terrier, Sydney silky",
    202: "soft-coated wheaten terrier",
    203: "West Highland white terrier",
    204: "Lhasa, Lhasa apso",
    205: "flat-coated retriever",
    206: "curly-coated retriever",
    207: "golden retriever",
    208: "Labrador retriever",
    209: "Chesapeake Bay retriever",
    210: "German short-haired pointer",
    211: "vizsla, Hungarian pointer",
    212: "English setter",
    213: "Irish setter, red setter",
    214: "Gordon setter",
    215: "Brittany spaniel",
    216: "clumber, clumber spaniel",
    217: "English springer, English springer spaniel",
    218: "Welsh springer spaniel",
    219: "cocker spaniel, English cocker spaniel, cocker",
    220: "Sussex spaniel",
    221: "Irish water spaniel",
    222: "kuvasz",
    223: "schipperke",
    224: "groenendael",
    225: "malinois",
    226: "briard",
    227: "kelpie",
    228: "komondor",
    229: "Old English sheepdog, bobtail",
    230: "Shetland sheepdog, Shetland sheep dog, Shetland",
    231: "collie",
    232: "Border collie",
    233: "Bouvier des Flandres, Bouviers des Flandres",
    234: "Rottweiler",
    235: "German shepherd, German shepherd dog, German police dog, alsatian",
    236: "Doberman, Doberman pinscher",
    237: "miniature pinscher",
    238: "Greater Swiss Mountain dog",
    239: "Bernese mountain dog",
    240: "Appenzeller",
    241: "EntleBucher",
    242: "boxer",
    243: "bull mastiff",
    244: "Tibetan mastiff",
    245: "French bulldog",
    246: "Great Dane",
    247: "Saint Bernard, St Bernard",
    248: "Eskimo dog, husky",
    249: "malamute, malemute, Alaskan malamute",
    250: "Siberian husky",
    251: "dalmatian, coach dog, carriage dog",
    252: "affenpinscher, monkey pinscher, monkey dog",
    253: "basenji",
    254: "pug, pug-dog",
    255: "Leonberg",
    256: "Newfoundland, Newfoundland dog",
    257: "Great Pyrenees",
    258: "Samoyed, Samoyede",
    259: "Pomeranian",
    260: "chow, chow chow",
    261: "keeshond",
    262: "Brabancon griffon",
    263: "Pembroke, Pembroke Welsh corgi",
    264: "Cardigan, Cardigan Welsh corgi",
    265: "toy poodle",
    266: "miniature poodle",
    267: "standard poodle",
    268: "Mexican hairless",
    269: "timber wolf, grey wolf, gray wolf, Canis lupus",
    270: "white wolf, Arctic wolf, Canis lupus tundrarum",
    271: "red wolf, maned wolf, Canis rufus, Canis niger",
    272: "coyote, prairie wolf, brush wolf, Canis latrans",
    273: "dingo, warrigal, warragal, Canis dingo",
    274: "dhole, Cuon alpinus",
    275: "African hunting dog, hyena dog, Cape hunting dog, Lycaon pictus",
    276: "hyena, hyaena",
    277: "red fox, Vulpes vulpes",
    278: "kit fox, Vulpes macrotis",
    279: "Arctic fox, white fox, Alopex lagopus",
    280: "grey fox, gray fox, Urocyon cinereoargenteus",
    281: "tabby, tabby cat",
    282: "tiger cat",
    283: "Persian cat",
    284: "Siamese cat, Siamese",
    285: "Egyptian cat",
    286: "cougar, puma, catamount, mountain lion, painter, " "panther, Felis concolor",
    287: "lynx, catamount",
    288: "leopard, Panthera pardus",
    289: "snow leopard, ounce, Panthera uncia",
    290: "jaguar, panther, Panthera onca, Felis onca",
    291: "lion, king of beasts, Panthera leo",
    292: "tiger, Panthera tigris",
    293: "cheetah, chetah, Acinonyx jubatus",
    294: "brown bear, bruin, Ursus arctos",
    295: "American black bear, black bear, Ursus americanus, " "Euarctos americanus",
    296: "ice bear, polar bear, Ursus Maritimus, Thalarctos maritimus",
    297: "sloth bear, Melursus ursinus, Ursus ursinus",
    298: "mongoose",
    299: "meerkat, mierkat",
    300: "tiger beetle",
    301: "ladybug, ladybeetle, lady beetle, ladybird, ladybird beetle",
    302: "ground beetle, carabid beetle",
    303: "long-horned beetle, longicorn, longicorn beetle",
    304: "leaf beetle, chrysomelid",
    305: "dung beetle",
    306: "rhinoceros beetle",
    307: "weevil",
    308: "fly",
    309: "bee",
    310: "ant, emmet, pismire",
    311: "grasshopper, hopper",
    312: "cricket",
    313: "walking stick, walkingstick, stick insect",
    314: "cockroach, roach",
    315: "mantis, mantid",
    316: "cicada, cicala",
    317: "leafhopper",
    318: "lacewing, lacewing fly",
    319: "dragonfly, darning needle, devil's darning needle, "
    "sewing needle, snake feeder, snake doctor, mosquito "
    "hawk, skeeter hawk",
    320: "damselfly",
    321: "admiral",
    322: "ringlet, ringlet butterfly",
    323: "monarch, monarch butterfly, milkweed butterfly, Danaus plexippus",
    324: "cabbage butterfly",
    325: "sulphur butterfly, sulfur butterfly",
    326: "lycaenid, lycaenid butterfly",
    327: "starfish, sea star",
    328: "sea urchin",
    329: "sea cucumber, holothurian",
    330: "wood rabbit, cottontail, cottontail rabbit",
    331: "hare",
    332: "Angora, Angora rabbit",
    333: "hamster",
    334: "porcupine, hedgehog",
    335: "fox squirrel, eastern fox squirrel, Sciurus niger",
    336: "marmot",
    337: "beaver",
    338: "guinea pig, Cavia cobaya",
    339: "sorrel",
    340: "zebra",
    341: "hog, pig, grunter, squealer, Sus scrofa",
    342: "wild boar, boar, Sus scrofa",
    343: "warthog",
    344: "hippopotamus, hippo, river horse, Hippopotamus amphibius",
    345: "ox",
    346: "water buffalo, water ox, Asiatic buffalo, Bubalus bubalis",
    347: "bison",
    348: "ram, tup",
    349: "bighorn, bighorn sheep, cimarron, Rocky Mountain "
    "bighorn, Rocky Mountain sheep, Ovis canadensis",
    350: "ibex, Capra ibex",
    351: "hartebeest",
    352: "impala, Aepyceros melampus",
    353: "gazelle",
    354: "Arabian camel, dromedary, Camelus dromedarius",
    355: "llama",
    356: "weasel",
    357: "mink",
    358: "polecat, fitch, foulmart, foumart, Mustela putorius",
    359: "black-footed ferret, ferret, Mustela nigripes",
    360: "otter",
    361: "skunk, polecat, wood pussy",
    362: "badger",
    363: "armadillo",
    364: "three-toed sloth, ai, Bradypus tridactylus",
    365: "orangutan, orang, orangutang, Pongo pygmaeus",
    366: "gorilla, Gorilla gorilla",
    367: "chimpanzee, chimp, Pan troglodytes",
    368: "gibbon, Hylobates lar",
    369: "siamang, Hylobates syndactylus, Symphalangus syndactylus",
    370: "guenon, guenon monkey",
    371: "patas, hussar monkey, Erythrocebus patas",
    372: "baboon",
    373: "macaque",
    374: "langur",
    375: "colobus, colobus monkey",
    376: "proboscis monkey, Nasalis larvatus",
    377: "marmoset",
    378: "capuchin, ringtail, Cebus capucinus",
    379: "howler monkey, howler",
    380: "titi, titi monkey",
    381: "spider monkey, Ateles geoffroyi",
    382: "squirrel monkey, Saimiri sciureus",
    383: "Madagascar cat, ring-tailed lemur, Lemur catta",
    384: "indri, indris, Indri indri, Indri brevicaudatus",
    385: "Indian elephant, Elephas maximus",
    386: "African elephant, Loxodonta africana",
    387: "lesser panda, red panda, panda, bear cat, cat bear, Ailurus fulgens",
    388: "giant panda, panda, panda bear, coon bear, Ailuropoda melanoleuca",
    389: "barracouta, snoek",
    390: "eel",
    391: "coho, cohoe, coho salmon, blue jack, silver salmon, " "Oncorhynchus kisutch",
    392: "rock beauty, Holocanthus tricolor",
    393: "anemone fish",
    394: "sturgeon",
    395: "gar, garfish, garpike, billfish, Lepisosteus osseus",
    396: "lionfish",
    397: "puffer, pufferfish, blowfish, globefish",
    398: "abacus",
    399: "abaya",
    400: "academic gown, academic robe, judge's robe",
    401: "accordion, piano accordion, squeeze box",
    402: "acoustic guitar",
    403: "aircraft carrier, carrier, flattop, attack aircraft carrier",
    404: "airliner",
    405: "airship, dirigible",
    406: "altar",
    407: "ambulance",
    408: "amphibian, amphibious vehicle",
    409: "analog clock",
    410: "apiary, bee house",
    411: "apron",
    412: "ashcan, trash can, garbage can, wastebin, ash bin, "
    "ash-bin, ashbin, dustbin, trash barrel, trash bin",
    413: "assault rifle, assault gun",
    414: "backpack, back pack, knapsack, packsack, rucksack, haversack",
    415: "bakery, bakeshop, bakehouse",
    416: "balance beam, beam",
    417: "balloon",
    418: "ballpoint, ballpoint pen, ballpen, Biro",
    419: "Band Aid",
    420: "banjo",
    421: "bannister, banister, balustrade, balusters, handrail",
    422: "barbell",
    423: "barber chair",
    424: "barbershop",
    425: "barn",
    426: "barometer",
    427: "barrel, cask",
    428: "barrow, garden cart, lawn cart, wheelbarrow",
    429: "baseball",
    430: "basketball",
    431: "bassinet",
    432: "bassoon",
    433: "bathing cap, swimming cap",
    434: "bath towel",
    435: "bathtub, bathing tub, bath, tub",
    436: "beach wagon, station wagon, wagon, estate car, "
    "beach waggon, station waggon, waggon",
    437: "beacon, lighthouse, beacon light, pharos",
    438: "beaker",
    439: "bearskin, busby, shako",
    440: "beer bottle",
    441: "beer glass",
    442: "bell cote, bell cot",
    443: "bib",
    444: "bicycle-built-for-two, tandem bicycle, tandem",
    445: "bikini, two-piece",
    446: "binder, ring-binder",
    447: "binoculars, field glasses, opera glasses",
    448: "birdhouse",
    449: "boathouse",
    450: "bobsled, bobsleigh, bob",
    451: "bolo tie, bolo, bola tie, bola",
    452: "bonnet, poke bonnet",
    453: "bookcase",
    454: "bookshop, bookstore, bookstall",
    455: "bottlecap",
    456: "bow",
    457: "bow tie, bow-tie, bowtie",
    458: "brass, memorial tablet, plaque",
    459: "brassiere, bra, bandeau",
    460: "breakwater, groin, groyne, mole, bulwark, seawall, jetty",
    461: "breastplate, aegis, egis",
    462: "broom",
    463: "bucket, pail",
    464: "buckle",
    465: "bulletproof vest",
    466: "bullet train, bullet",
    467: "butcher shop, meat market",
    468: "cab, hack, taxi, taxicab",
    469: "caldron, cauldron",
    470: "candle, taper, wax light",
    471: "cannon",
    472: "canoe",
    473: "can opener, tin opener",
    474: "cardigan",
    475: "car mirror",
    476: "carousel, carrousel, merry-go-round, roundabout, whirligig",
    477: "carpenter's kit, tool kit",
    478: "carton",
    479: "car wheel",
    480: "cash machine, cash dispenser, automated teller "
    "machine, automatic teller machine, automated teller, "
    "automatic teller, ATM",
    481: "cassette",
    482: "cassette player",
    483: "castle",
    484: "catamaran",
    485: "CD player",
    486: "cello, violoncello",
    487: "cellular telephone, cellular phone, cellphone, cell, mobile phone",
    488: "chain",
    489: "chainlink fence",
    490: "chain mail, ring mail, mail, chain armor, chain "
    "armour, ring armor, ring armour",
    491: "chain saw, chainsaw",
    492: "chest",
    493: "chiffonier, commode",
    494: "chime, bell, gong",
    495: "china cabinet, china closet",
    496: "Christmas stocking",
    497: "church, church building",
    498: "cinema, movie theater, movie theatre, movie house, picture palace",
    499: "cleaver, meat cleaver, chopper",
    500: "cliff dwelling",
    501: "cloak",
    502: "clog, geta, patten, sabot",
    503: "cocktail shaker",
    504: "coffee mug",
    505: "coffeepot",
    506: "coil, spiral, volute, whorl, helix",
    507: "combination lock",
    508: "computer keyboard, keypad",
    509: "confectionery, confectionary, candy store",
    510: "container ship, containership, container vessel",
    511: "convertible",
    512: "corkscrew, bottle screw",
    513: "cornet, horn, trumpet, trump",
    514: "cowboy boot",
    515: "cowboy hat, ten-gallon hat",
    516: "cradle",
    517: "crane",
    518: "crash helmet",
    519: "crate",
    520: "crib, cot",
    521: "Crock Pot",
    522: "croquet ball",
    523: "crutch",
    524: "cuirass",
    525: "dam, dike, dyke",
    526: "desk",
    527: "desktop computer",
    528: "dial telephone, dial phone",
    529: "diaper, nappy, napkin",
    530: "digital clock",
    531: "digital watch",
    532: "dining table, board",
    533: "dishrag, dishcloth",
    534: "dishwasher, dish washer, dishwashing machine",
    535: "disk brake, disc brake",
    536: "dock, dockage, docking facility",
    537: "dogsled, dog sled, dog sleigh",
    538: "dome",
    539: "doormat, welcome mat",
    540: "drilling platform, offshore rig",
    541: "drum, membranophone, tympan",
    542: "drumstick",
    543: "dumbbell",
    544: "Dutch oven",
    545: "electric fan, blower",
    546: "electric guitar",
    547: "electric locomotive",
    548: "entertainment center",
    549: "envelope",
    550: "espresso maker",
    551: "face powder",
    552: "feather boa, boa",
    553: "file, file cabinet, filing cabinet",
    554: "fireboat",
    555: "fire engine, fire truck",
    556: "fire screen, fireguard",
    557: "flagpole, flagstaff",
    558: "flute, transverse flute",
    559: "folding chair",
    560: "football helmet",
    561: "forklift",
    562: "fountain",
    563: "fountain pen",
    564: "four-poster",
    565: "freight car",
    566: "French horn, horn",
    567: "frying pan, frypan, skillet",
    568: "fur coat",
    569: "garbage truck, dustcart",
    570: "gasmask, respirator, gas helmet",
    571: "gas pump, gasoline pump, petrol pump, island dispenser",
    572: "goblet",
    573: "go-kart",
    574: "golf ball",
    575: "golfcart, golf cart",
    576: "gondola",
    577: "gong, tam-tam",
    578: "gown",
    579: "grand piano, grand",
    580: "greenhouse, nursery, glasshouse",
    581: "grille, radiator grille",
    582: "grocery store, grocery, food market, market",
    583: "guillotine",
    584: "hair slide",
    585: "hair spray",
    586: "half track",
    587: "hammer",
    588: "hamper",
    589: "hand blower, blow dryer, blow drier, hair dryer, hair drier",
    590: "hand-held computer, hand-held microcomputer",
    591: "handkerchief, hankie, hanky, hankey",
    592: "hard disc, hard disk, fixed disk",
    593: "harmonica, mouth organ, harp, mouth harp",
    594: "harp",
    595: "harvester, reaper",
    596: "hatchet",
    597: "holster",
    598: "home theater, home theatre",
    599: "honeycomb",
    600: "hook, claw",
    601: "hoopskirt, crinoline",
    602: "horizontal bar, high bar",
    603: "horse cart, horse-cart",
    604: "hourglass",
    605: "iPod",
    606: "iron, smoothing iron",
    607: "jack-o'-lantern",
    608: "jean, blue jean, denim",
    609: "jeep, landrover",
    610: "jersey, T-shirt, tee shirt",
    611: "jigsaw puzzle",
    612: "jinrikisha, ricksha, rickshaw",
    613: "joystick",
    614: "kimono",
    615: "knee pad",
    616: "knot",
    617: "lab coat, laboratory coat",
    618: "ladle",
    619: "lampshade, lamp shade",
    620: "laptop, laptop computer",
    621: "lawn mower, mower",
    622: "lens cap, lens cover",
    623: "letter opener, paper knife, paperknife",
    624: "library",
    625: "lifeboat",
    626: "lighter, light, igniter, ignitor",
    627: "limousine, limo",
    628: "liner, ocean liner",
    629: "lipstick, lip rouge",
    630: "Loafer",
    631: "lotion",
    632: "loudspeaker, speaker, speaker unit, loudspeaker " "system, speaker system",
    633: "loupe, jeweler's loupe",
    634: "lumbermill, sawmill",
    635: "magnetic compass",
    636: "mailbag, postbag",
    637: "mailbox, letter box",
    638: "maillot",
    639: "maillot, tank suit",
    640: "manhole cover",
    641: "maraca",
    642: "marimba, xylophone",
    643: "mask",
    644: "matchstick",
    645: "maypole",
    646: "maze, labyrinth",
    647: "measuring cup",
    648: "medicine chest, medicine cabinet",
    649: "megalith, megalithic structure",
    650: "microphone, mike",
    651: "microwave, microwave oven",
    652: "military uniform",
    653: "milk can",
    654: "minibus",
    655: "miniskirt, mini",
    656: "minivan",
    657: "missile",
    658: "mitten",
    659: "mixing bowl",
    660: "mobile home, manufactured home",
    661: "Model T",
    662: "modem",
    663: "monastery",
    664: "monitor",
    665: "moped",
    666: "mortar",
    667: "mortarboard",
    668: "mosque",
    669: "mosquito net",
    670: "motor scooter, scooter",
    671: "mountain bike, all-terrain bike, off-roader",
    672: "mountain tent",
    673: "mouse, computer mouse",
    674: "mousetrap",
    675: "moving van",
    676: "muzzle",
    677: "nail",
    678: "neck brace",
    679: "necklace",
    680: "nipple",
    681: "notebook, notebook computer",
    682: "obelisk",
    683: "oboe, hautboy, hautbois",
    684: "ocarina, sweet potato",
    685: "odometer, hodometer, mileometer, milometer",
    686: "oil filter",
    687: "organ, pipe organ",
    688: "oscilloscope, scope, cathode-ray oscilloscope, CRO",
    689: "overskirt",
    690: "oxcart",
    691: "oxygen mask",
    692: "packet",
    693: "paddle, boat paddle",
    694: "paddlewheel, paddle wheel",
    695: "padlock",
    696: "paintbrush",
    697: "pajama, pyjama, pj's, jammies",
    698: "palace",
    699: "panpipe, pandean pipe, syrinx",
    700: "paper towel",
    701: "parachute, chute",
    702: "parallel bars, bars",
    703: "park bench",
    704: "parking meter",
    705: "passenger car, coach, carriage",
    706: "patio, terrace",
    707: "pay-phone, pay-station",
    708: "pedestal, plinth, footstall",
    709: "pencil box, pencil case",
    710: "pencil sharpener",
    711: "perfume, essence",
    712: "Petri dish",
    713: "photocopier",
    714: "pick, plectrum, plectron",
    715: "pickelhaube",
    716: "picket fence, paling",
    717: "pickup, pickup truck",
    718: "pier",
    719: "piggy bank, penny bank",
    720: "pill bottle",
    721: "pillow",
    722: "ping-pong ball",
    723: "pinwheel",
    724: "pirate, pirate ship",
    725: "pitcher, ewer",
    726: "plane, carpenter's plane, woodworking plane",
    727: "planetarium",
    728: "plastic bag",
    729: "plate rack",
    730: "plow, plough",
    731: "plunger, plumber's helper",
    732: "Polaroid camera, Polaroid Land camera",
    733: "pole",
    734: "police van, police wagon, paddy wagon, patrol wagon, " "wagon, black Maria",
    735: "poncho",
    736: "pool table, billiard table, snooker table",
    737: "pop bottle, soda bottle",
    738: "pot, flowerpot",
    739: "potter's wheel",
    740: "power drill",
    741: "prayer rug, prayer mat",
    742: "printer",
    743: "prison, prison house",
    744: "projectile, missile",
    745: "projector",
    746: "puck, hockey puck",
    747: "punching bag, punch bag, punching ball, punchball",
    748: "purse",
    749: "quill, quill pen",
    750: "quilt, comforter, comfort, puff",
    751: "racer, race car, racing car",
    752: "racket, racquet",
    753: "radiator",
    754: "radio, wireless",
    755: "radio telescope, radio reflector",
    756: "rain barrel",
    757: "recreational vehicle, RV, R.V.",
    758: "reel",
    759: "reflex camera",
    760: "refrigerator, icebox",
    761: "remote control, remote",
    762: "restaurant, eating house, eating place, eatery",
    763: "revolver, six-gun, six-shooter",
    764: "rifle",
    765: "rocking chair, rocker",
    766: "rotisserie",
    767: "rubber eraser, rubber, pencil eraser",
    768: "rugby ball",
    769: "rule, ruler",
    770: "running shoe",
    771: "safe",
    772: "safety pin",
    773: "saltshaker, salt shaker",
    774: "sandal",
    775: "sarong",
    776: "sax, saxophone",
    777: "scabbard",
    778: "scale, weighing machine",
    779: "school bus",
    780: "schooner",
    781: "scoreboard",
    782: "screen, CRT screen",
    783: "screw",
    784: "screwdriver",
    785: "seat belt, seatbelt",
    786: "sewing machine",
    787: "shield, buckler",
    788: "shoe shop, shoe-shop, shoe store",
    789: "shoji",
    790: "shopping basket",
    791: "shopping cart",
    792: "shovel",
    793: "shower cap",
    794: "shower curtain",
    795: "ski",
    796: "ski mask",
    797: "sleeping bag",
    798: "slide rule, slipstick",
    799: "sliding door",
    800: "slot, one-armed bandit",
    801: "snorkel",
    802: "snowmobile",
    803: "snowplow, snowplough",
    804: "soap dispenser",
    805: "soccer ball",
    806: "sock",
    807: "solar dish, solar collector, solar furnace",
    808: "sombrero",
    809: "soup bowl",
    810: "space bar",
    811: "space heater",
    812: "space shuttle",
    813: "spatula",
    814: "speedboat",
    815: "spider web, spider's web",
    816: "spindle",
    817: "sports car, sport car",
    818: "spotlight, spot",
    819: "stage",
    820: "steam locomotive",
    821: "steel arch bridge",
    822: "steel drum",
    823: "stethoscope",
    824: "stole",
    825: "stone wall",
    826: "stopwatch, stop watch",
    827: "stove",
    828: "strainer",
    829: "streetcar, tram, tramcar, trolley, trolley car",
    830: "stretcher",
    831: "studio couch, day bed",
    832: "stupa, tope",
    833: "submarine, pigboat, sub, U-boat",
    834: "suit, suit of clothes",
    835: "sundial",
    836: "sunglass",
    837: "sunglasses, dark glasses, shades",
    838: "sunscreen, sunblock, sun blocker",
    839: "suspension bridge",
    840: "swab, swob, mop",
    841: "sweatshirt",
    842: "swimming trunks, bathing trunks",
    843: "swing",
    844: "switch, electric switch, electrical switch",
    845: "syringe",
    846: "table lamp",
    847: "tank, army tank, armored combat vehicle, armoured combat vehicle",
    848: "tape player",
    849: "teapot",
    850: "teddy, teddy bear",
    851: "television, television system",
    852: "tennis ball",
    853: "thatch, thatched roof",
    854: "theater curtain, theatre curtain",
    855: "thimble",
    856: "thresher, thrasher, threshing machine",
    857: "throne",
    858: "tile roof",
    859: "toaster",
    860: "tobacco shop, tobacconist shop, tobacconist",
    861: "toilet seat",
    862: "torch",
    863: "totem pole",
    864: "tow truck, tow car, wrecker",
    865: "toyshop",
    866: "tractor",
    867: "trailer truck, tractor trailer, trucking rig, rig, "
    "articulated lorry, semi",
    868: "tray",
    869: "trench coat",
    870: "tricycle, trike, velocipede",
    871: "trimaran",
    872: "tripod",
    873: "triumphal arch",
    874: "trolleybus, trolley coach, trackless trolley",
    875: "trombone",
    876: "tub, vat",
    877: "turnstile",
    878: "typewriter keyboard",
    879: "umbrella",
    880: "unicycle, monocycle",
    881: "upright, upright piano",
    882: "vacuum, vacuum cleaner",
    883: "vase",
    884: "vault",
    885: "velvet",
    886: "vending machine",
    887: "vestment",
    888: "viaduct",
    889: "violin, fiddle",
    890: "volleyball",
    891: "waffle iron",
    892: "wall clock",
    893: "wallet, billfold, notecase, pocketbook",
    894: "wardrobe, closet, press",
    895: "warplane, military plane",
    896: "washbasin, handbasin, washbowl, lavabo, wash-hand basin",
    897: "washer, automatic washer, washing machine",
    898: "water bottle",
    899: "water jug",
    900: "water tower",
    901: "whiskey jug",
    902: "whistle",
    903: "wig",
    904: "window screen",
    905: "window shade",
    906: "Windsor tie",
    907: "wine bottle",
    908: "wing",
    909: "wok",
    910: "wooden spoon",
    911: "wool, woolen, woollen",
    912: "worm fence, snake fence, snake-rail fence, Virginia fence",
    913: "wreck",
    914: "yawl",
    915: "yurt",
    916: "web site, website, internet site, site",
    917: "comic book",
    918: "crossword puzzle, crossword",
    919: "street sign",
    920: "traffic light, traffic signal, stoplight",
    921: "book jacket, dust cover, dust jacket, dust wrapper",
    922: "menu",
    923: "plate",
    924: "guacamole",
    925: "consomme",
    926: "hot pot, hotpot",
    927: "trifle",
    928: "ice cream, icecream",
    929: "ice lolly, lolly, lollipop, popsicle",
    930: "French loaf",
    931: "bagel, beigel",
    932: "pretzel",
    933: "cheeseburger",
    934: "hotdog, hot dog, red hot",
    935: "mashed potato",
    936: "head cabbage",
    937: "broccoli",
    938: "cauliflower",
    939: "zucchini, courgette",
    940: "spaghetti squash",
    941: "acorn squash",
    942: "butternut squash",
    943: "cucumber, cuke",
    944: "artichoke, globe artichoke",
    945: "bell pepper",
    946: "cardoon",
    947: "mushroom",
    948: "Granny Smith",
    949: "strawberry",
    950: "orange",
    951: "lemon",
    952: "fig",
    953: "pineapple, ananas",
    954: "banana",
    955: "jackfruit, jak, jack",
    956: "custard apple",
    957: "pomegranate",
    958: "hay",
    959: "carbonara",
    960: "chocolate sauce, chocolate syrup",
    961: "dough",
    962: "meat loaf, meatloaf",
    963: "pizza, pizza pie",
    964: "potpie",
    965: "burrito",
    966: "red wine",
    967: "espresso",
    968: "cup",
    969: "eggnog",
    970: "alp",
    971: "bubble",
    972: "cliff, drop, drop-off",
    973: "coral reef",
    974: "geyser",
    975: "lakeside, lakeshore",
    976: "promontory, headland, head, foreland",
    977: "sandbar, sand bar",
    978: "seashore, coast, seacoast, sea-coast",
    979: "valley, vale",
    980: "volcano",
    981: "ballplayer, baseball player",
    982: "groom, bridegroom",
    983: "scuba diver",
    984: "rapeseed",
    985: "daisy",
    986: "yellow lady's slipper, yellow lady-slipper, Cypripedium "
    "calceolus, Cypripedium parviflorum",
    987: "corn",
    988: "acorn",
    989: "hip, rose hip, rosehip",
    990: "buckeye, horse chestnut, conker",
    991: "coral fungus",
    992: "agaric",
    993: "gyromitra",
    994: "stinkhorn, carrion fungus",
    995: "earthstar",
    996: "hen-of-the-woods, hen of the woods, Polyporus frondosus, " "Grifola frondosa",
    997: "bolete",
    998: "ear, spike, capitulum",
    999: "toilet tissue, toilet paper, bathroom tissue",
}
